//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "error.h"
#include "alphabet.h"
#include "str.h"

/**
 * A node of the prefix tree storing a set of words. It represents one
 * string corresponding to the path from the root to the current
 * node. Additionally, the node contains the canonical representative
 * of the string 'str' and a unique identifier of the string 'str_id'.
 */
struct Alphabet::Node
{
   /**
    * An array of pointers to child nodes. There are 256 pointers, one
    * for every possible character.
    */
   Vector<Alphabet::Node*> child;

   /**
    * The string associated with the node. 'NULL' if the node is
    * transient i.e., it is accessible with a string that does not
    * belong to the alphabet.
    */
   String str;

   /**
    * A unique identifier of a string. Has value different from '-1'
    * iff 'str' is different from 'NULL'.
    */
   int   str_id;

   /**
    * Standard constructor. Initialises the fields and in particular
    * the 'child' array is allocated and all its cells set to 'NULL';
    */
   Node();

   /**
    * Standard destructor. Deallocates the resources used to represent
    * the prefix tree. Calls the destructor recursively to all the
    * children. Does not release the memory used to store 'str' and if
    * 'str' is a result of duplication, then the memory should be
    * released elsewhere.
    */
   ~Node();

   /**
    * Deep cloning of the node. Called recursively on the
    * children.
    */
   Node* clone();

   /**
    * Attempts to find the node reached by a path corresponding to the
    * input string (evaluated from the current node). Returns 'NULL'
    * if there is no such node. Typically, called from the root node
    * of the prefix tree.
    */
   Node* find(String& str);

   /**
    * Finds the node reached by a path corresponding to the input
    * string. If the node does not exists in the prefix tree, it is
    * added (including all nonexistent ancestor nodes). Typically,
    * called from the root node of the prefix tree.
    */
   Node* find_or_add(String& str);
   
   /** 
    *  Adds strings stored in the prefix tree to the vector. If the
    *  node stores a string in 'str' (i.e., 'str' is non-null), then
    *  'str' is added to the output vector at position 'str_id'. No
    *  string is added if 'str' is 'NULL'. Calls recursively for every
    *  descendant node. The uniqueness of 'str_id' values ensures that
    *  all strings are added to the string vector but the caller must
    *  ensure that the vector size is sufficient.
    */ 
   void  pack_into(Vector<String>& store);
};

Alphabet::Node::Node() 
   : child(256)
{
   for (int i = 0; i < 256; i++) {
      this->child[i] = NULL;
   }
   
   this->str = String::Null;
   this->str_id = -1;
}

Alphabet::Node::~Node()
{
   for (int i = 0; i < 256; i++) {
      if (child[i] != NULL) {
         delete child[i];
      }
   }
}

Alphabet::Node* Alphabet::Node::clone()
{
   Node* n = new Node();
   
   for (int i = 0; i < 256; i++) {
      if (this->child[i] != NULL) {
         n->child[i] = this->child[i]->clone();
      }
   }

   if (! this->str.null()) {
      n->str_id = this->str_id;
      n->str = this->str;
   }

   return n;
}


Alphabet::Node* Alphabet::Node::find(String& str)
{
   ASSERT(! str.null());
   
   Node* tmp = this;
   unsigned char* p = (unsigned char*) str.str;
   
   while (*p != '\0') {
      if (tmp->child[*p] == NULL) {
         return NULL;
      }

      tmp = tmp->child[*p];
      p++;
   }
   
   return tmp;
}

Alphabet::Node* Alphabet::Node::find_or_add(String& str)
{
   ASSERT(!str.null());
   
   Alphabet::Node* tmp = this;
   unsigned char* p = (unsigned char*) str.str;
   
   while (*p != '\0') {
      if (tmp->child[*p] == NULL) {
         tmp->child[*p] = new Alphabet::Node();
      }

      tmp = tmp->child[*p];
      p++;
   }
   
   return tmp;
}

void Alphabet::Node::pack_into(Vector<String>& store)
{
   if (!this->str.null()) {
      store[this->str_id] = this->str;
   }
   
   for (int i = 0; i < 256; i++) {
      if (this->child[i] != NULL) {
         this->child[i]->pack_into(store);
      }
   }
}

Alphabet::Alphabet()
{
   this->root = new Node();
   this->frozen = false;
   this->size = 0;
   this->ref_count = 1;
}

Alphabet::Alphabet(Alphabet* alphabet)
{
   ASSERT(alphabet != NULL);

   this->root = alphabet->root->clone();
   this->frozen = false;
   this->size = alphabet->size;
   this->ref_count = 1;
}

Alphabet::~Alphabet()
{
   delete this->root;
}

Alphabet* Alphabet::retain()
{
   this->ref_count++;
   return this;
}

void Alphabet::release()
{
   this->ref_count--;
   if (this->ref_count == 0) {
      delete this;
   }
}

Alphabet::Node* Alphabet::find_node(String& str)
{
   if (this->frozen) {
      return this->root->find(str);
   } else {
      Alphabet::Node* n = this->root->find_or_add(str);

      if (n->str.null()) {
         n->str = str;
         n->str_id = this->size++;
      }

      return n;
   }
}

String& Alphabet::find(String& str)
{
   ASSERT(! str.null());
   Node* n = find_node(str);
   
   if (n != NULL) {
      return n->str;
   } else {
      return String::Null;
   }
}

bool Alphabet::contains(String& str)
{
   ASSERT(!str.null());

   return this->root->find(str) != NULL;
}

int Alphabet::find_id(String& str)
{
   ASSERT(! str.null());
   Node* n = find_node(str);
   
   if (n != NULL) {
      return n->str_id;
   } else {
      return -1;
   }
}

void Alphabet::freeze()
{
   this->frozen = true;
}

void Alphabet::pack_into(Vector<String>& store)
{
   if (store.null() || store.size != this->size) {
      store.reinit(this->size);
   }

   this->root->pack_into(store);
}

void Alphabet::translate(Vector<String>& str, Vector<int>& id)
{
   ASSERT(! str.null());

   if (id.null() || id.size != str.size) {
      id.reinit(str.size);
   }

   for (int i = 0; i < str.size; i++) {
      ASSERT(!str[i].null());
      id[i] = this->find_id(str[i]);
   }
}

#include <stdio.h>

void Alphabet::print()
{
   Vector<String> tmp;
   
   pack_into(tmp);

   printf("{");
   for (int i = 0; i < tmp.size; i++) {
      printf("%i : ",find_id(tmp[i]));
      printf("%s",tmp[i].str);
      if (i < size - 1) {
         printf(", ");
      }
   }
   printf("}");   
}

#include "utils.h"
#include <stdio.h>
void Alphabet::print(char* name, int ind, bool newline = true)
{
   indent(ind);
   Vector<String> tmp;
   
   pack_into(tmp);

   if (name != NULL) {
      printf("%s = ", name);
   }
   printf("{");
   for (int i = 0; i < size; i++) {
      printf("%s : %i",tmp[i].str,find_id(tmp[i]));
      if (i < size - 1) {
         printf(", ");
      }
   }
   printf("}");
   if (newline) {
      printf("\n");
   }
}

void Alphabet::scan_str(String s, Vector<int>& out)
{
   List<int> out_l;
   String s_copy(s.str);
   char* curr = s_copy.str;

   while (true) {
      //skip white-space
      while (*curr != '\0' && isspace(*curr)) {
         curr++;
      }

      if (*curr == '\0') {
         break;
      }

      if (! isalpha(*curr)) {
         ERROR("Wrong character '%c' (code=%i) at position %li.",
               *curr,*curr,curr-s_copy.str);
      }

      char* start = curr;
      while (isalnum(*curr) || *curr == '-') {
         curr++;
      }

      if (!(isspace(*curr) || *curr == '\0')) {
         ERROR("Wrong character '%c' (code=%i) at position %li.",
               *curr,*curr,curr-s_copy.str);
      }

      if (*curr != '\0') {
         *curr = '\0';
         curr++;
      }

      String str(start,true);
      out_l.add(this->find_id(str));
   }

   out.reinit(out_l);
}

void Alphabet::scan_file(const char* file_name, Vector<int>& out)
{
   static char buffer[8*1024];
   List<int> out_l;

   FILE* file = fopen(file_name, "r");

   if (file == NULL) {
      ERROR("Error opening the file '%s' for lecture.", file_name);
   }

   int curr;
   int line = 1;
   int column = 1;

   while (true) {

      // skip white-space
      while ((curr = fgetc(file)) != -1 && isspace(curr)) {
         column++;
         if (curr == '\n') {
            column = 0;
            line++;
         }
      }

      if (curr == -1) {
         if (ferror(file) != 0) {
            ERROR("Error reading file '%s'.",file_name);
         }
         if (feof(file) != 0) {
            break;
         }
      }

      if (! isalpha(curr)) {
         ERROR("Wrong character '%c' (code=%i) at line %i column %i.",
               curr,curr,line,column);
      }

      char* tmp = buffer;
      while (isalnum(curr) || curr == '-') {
         *tmp = curr;
         tmp++;
         curr = fgetc(file);

         column++;
         if (curr == '\n') {
            column = 0;
            line++;
         }
      }
      *tmp = '\0';
      String str(buffer,true);
      out_l.add(this->find_id(str));
   }
   out.reinit(out_l);
}

