//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_ALPHABET_H_
#define LiRE_ALPHABET_H_

#include "vector.h"
#include "str.h"

/**
 * A collection of strings with unique identifiers.
 *
 * This structure stores the strings using a prefix tree. The prefix
 * tree is also used to associate with every stored string a unique
 * identifier, a natural number from `0' to size-1. 
 * 
 * Function 'find' is used to add a new string to the alphabet and to
 * retrieve the string stored in the alphabet that represents the
 * given input string (if the input string is already contained in the
 * alphabet or copying is enabled). This allows a basic string
 * canonization scheme: 
 * 
 * \code
 * Alphabet* a = new Alphabet();
 * 
 * String s1 = "abc";     
 * String s2 = strdup(s1);  
 * 
 * ASSERT(s1 == s2); // 's1' and 's2' contain the same string 
 * ASSERT(s1 != s2); // but in different memory locations
 * 
 * ASSERT(a->find(s1) == a->find(s2));
 * \endcode
 *
 * An alphabet may be frozen by calling function 'freeze' to prevent
 * adding further string, and then the 'frozen' flag is set to
 * true. When the alphabet is frozen a call to 'find' returns 'NULL'
 * if the alphabet does not contain the input string.
 *
 * Function 'contains' checks if the input string is present in the
 * alphabet (without attempting to add it).
 * 
 * Function 'find_id' returns a unique identifier associated with the
 * input string. The identifiers range from '0' to 'size-1' and
 * correspond to the order in which the strings have been added to the
 * dictionary. Similarly to 'find', the function 'find_id' adds the
 * input string to the alphabet if it is not already present and the
 * alphabet is not frozen. If the alphabet is frozen and the input
 * string is not recognized by the alphabet, '-1' is returned.
 */
struct Alphabet
{
   /**
    * A node of the prefix tree storing a set of words.
    */
   struct Node;

   /**
    * The root node of the prefix tree. 
    */
   Node* root;
   /**
    * The number of strings stored in the alphabet. 
    */
   int   size;
   /**
    * Indicates whether the dictionary is frozen. If it is, then no
    * new strings can be added.
    */
   bool  frozen; 

   /** 
    * Reference counter. Counts the number of active references to the
    * given alphabet. Reference counting is a memory management
    * technique that allows a group of owner objects to manipulate on
    * the same target object without the negative consequences of the
    * target object being destroyed by one owner while other owner may
    * still require to access the target object. The method 'retain'
    * is used to increase the counter and should be any object that
    * wishes to own the target object. The method 'release' is used to
    * decrease the counter and should be used instead of the 'delete'
    * operator (which is called automatically once the counter reaches
    * '0'). The initial value of the counter is '1'. 
    */
   int   ref_count; 

   /**
    * Standard Constructor. Constructs an empty alphabet.
    */
   Alphabet();

   /**
    * Constructor coping the contents of a given dictionary. The
    * copying procedure works in time linear in the the size of the
    * prefix tree representing the word set of the input alphabet.
    */
   Alphabet(Alphabet* dict);

   /**
    * Standard destructor.
    */
   ~Alphabet();

   /**
    * Increases the reference counter. Returns the pointer to the
    * current alphabet.
    */
   Alphabet* retain();

   /**
    * Decreases the reference counter. If the reference counter
    * reaches '0', the alphabet is destroyed. 
    */
   void release();

   /**
    * Finds the representative of the input string. Returns null
    * string if the alphabet is frozen and does not contain the input
    * string. If the input string is not in the alphabet, the alphabet
    * is not frozen and does not duplicate the input string, then the
    * output string is exactly the same pointer as the input. The
    * input string must be nonempty.
    *
    * Time: O(|s|).
    */
   String& find(String& s);
   
   /**
    * Checks if the dictionary contains the given string. 
    *
    * Time: O(|s|).
    */
   bool contains(String& s);

   /**
    * Returns a unique id associated with the given string. Returns
    * '-1' if the alphabet is frozen and the given string is unknown. 
    *
    * Time: O(|s|).
    */
   int find_id(String& str);

   /**
    * Disables adding new strings to the alphabet. 
    * 
    * Time: O(1).
    */
   void  freeze();

   /**
    * Places all strings into the given string vector. Reinitializes
    * the vector if needed. The time required for this operation is
    * linear in the size of the prefix tree used to store the string
    * of the alphabet.
    */
   void pack_into(Vector<String>& store);

   /**
    * Translates a vector of strings into a vector of unique
    * identifiers. Iteratively calls the 'find_id' method on
    * subsequent elements of the input string vector.  
    */
   void translate(Vector<String>& s, Vector<int>& id);

   /**
    * Parses the input string.
    */
   void scan_str(String s, Vector<int>& out);
   /**
    * Parses the given file.
    */
   void scan_file(const char* file_name, Vector<int>& out);

   // For debugging
   void print();
   void print(char* name, int ind, bool newline);
   void print(char* name, bool newline = true) {print(name,0,newline);};

   /**
    * Finds the node of the prefix tree storing the given string. For
    * private use only.
    */
   Node* find_node(String& str);

    // -----------------------------------------------------------------------
    //  Unimplemented constructors and operators
    // -----------------------------------------------------------------------
    Alphabet(const Alphabet&);
    Alphabet& operator=(const Alphabet&);
};

#endif // LiRE_ALPHABET_H_
