//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_ARRAY_H_
#define LiRE_ARRAY_H_

#include <tr1/unordered_map>

#include "error.h"
#include "utils.h"
#include "logger.h"

/**
 * Associative array of fixed size with constant initialization time.
 */

template<class T> 
struct Array
{
   /**
    * Size (cell count) of the array. 
    */
   int  size;   
   /**
    * Array storing the proper values.  
    */
   T*   value;  
   /**
    * The count of cells initialized so far. 
    */
   uint  count; 
   /**
    * The index for initialized cells. 'index[i]' is the index of
    * the 'i'-th initialized cell.
    */
   uint* index;
   /**
    * Reverse index for initialized cells. 'value[i]' is initialized
    * iff 'init_check[i] < count' and 'index[init_check[i]] == i'.
    */ 
   uint* init_check; 

   /**
    * Standard c'tor allocates the requested amount of memory.
    */
   Array(int size);

   /**
    * Standard d'tor frees all allocated memory.
    */
   ~Array();
   
   /**
    * Checks if the 'i'-th cell has been initialized before.
    */
   bool is_null(int i); 
   
   /**
    * Initializes the i-th cell to the default value for T (NULL for
    * pointers). For private use only.
    */
   void init(int i);   

   /**
    * Gives the reference to the 'i'-th cell. The cell is initialized
    * to the default value for 'T' if it hasn't been accessed before.
    */
   T& operator[](const int i);

   // Disable copy and assignment;
   void operator=(const Array& l);
   Array(const Array& v);
};

template<class T>
Array<T>::Array(int size)
{
   this->size = size;
   this->count = 0;
   this->value = new T[size];
   this->init_check = new uint[size];
   this->index = new uint[size];
}

template<class T> 
Array<T>::~Array()
{
   delete[] this->value;
   delete[] this->init_check;
   delete[] this->index;
}

template<class T> 
bool Array<T>::is_null(int i)
{
   ASSERT(0 <= i && i < this->size);

   return ! (this->init_check[i] < this->count && 
             this->index[this->init_check[i]] == static_cast<uint>(i));
}

template<class T> 
void Array<T>::init(int i)
{
   ASSERT(0 <= i && i < this->size);
   ASSERT(is_null(i));

   this->value[i] = NULL;
   this->init_check[i] = this->count;
   this->index[this->count] = i;
   this->count++;
}

template<class T> 
T& Array<T>::operator[](const int i)
{
   ASSERT(0 <= i && i < this->size);

   if (is_null(i)) {
      init(i);
   }

   return this->value[i];
}

/**
 * An associative array based on hashtables.
 */
template<class T>
struct HArray
{
   std::tr1::unordered_map<int,T> ht;
   int size;
   HArray();
   HArray(int size);
   ~HArray();
   T& operator[](const int i);

   // Disable copy and assignment;
   void operator=(const HArray& l);
   HArray(const HArray& v);
};

template<class T>
HArray<T>::HArray(int size)
{
   this->size = size;
}

template<class T>
HArray<T>::HArray()
{
   this->size = -1;
}

template<class T>
HArray<T>::~HArray()
{
}

template<class T>
T& HArray<T>::operator[](const int i)
{
   ASSERT(size = -1 ||(0 <= i && i < size));
   return ht[i];
}

#endif // LiRE_ARRAY_H_
