//
// This file is part of Chipmunk, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// Chipmunk is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, version 3 of the License, or any later
// version.
//
// Chipmunk is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License
// along with Chipmunk. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef RE_COLLECTIONS_H_
#define RE_COLLECTIONS_H_

#include <stdio.h>
#include <stdlib.h>
#include <tr1/unordered_map>



#include "error.h"
#include "utils.h"
#include "logger.h"

//
// Simple light-weight containers.
//


template<class T> struct Vector;
template<class T> struct SubVector;
template<class T> struct List;
template<class T> struct Queue;
template<class T> struct Stack;
template<class T> struct Array;
template<class T> struct HArray;


//
// VECTOR. Fixed-length array wrapper with reference counter and
// subvector facility.
//
template<class T> 
struct Vector 
{
   T*   res;
   int* ref_count;

   T*   a;
   int  size;

   Vector();
   Vector(int size);
   Vector(const Vector<T>& v);
   Vector(const Vector<T>& v, int offset, int size);
   Vector(const List<T>& l);

   ~Vector();

   void init();
   void init(int size);
   void init(const Vector<T>& l);
   void init(const Vector<T>& l, int offset, int size);
   void init(const List<T>& l);

   void reinit();
   void reinit(int size);
   void reinit(const Vector<T>& l);
   void reinit(const Vector<T>& l, int offset, int size);
   void reinit(const List<T>& l);

   void purge();

   bool null() const;
   bool empty() const;

   T& operator[](int i);

   // Disable assignment and copy;
   void operator=(const Vector& l);
   void operator=(const List<T>& l);
   void operator=(int size);
};

template<class T> 
Vector<T>::Vector()
{
   init();
}

template<class T> 
Vector<T>::Vector(int size)
{
   init(size);
}

template<class T> 
Vector<T>::Vector(const List<T>& l)
{
   init(l);
}

template<class T> 
Vector<T>::Vector(const Vector<T>& v)
{
   init(v);
}

template<class T> 
Vector<T>::Vector(const Vector<T>& v, int offset, int size)
{
   init(v,offset,size);
}

template<class T> 
Vector<T>::~Vector()
{
   this->purge();
}

template<class T> 
void Vector<T>::init() 
{
   this->res = NULL;
   this->ref_count = NULL;
   this->a = NULL;
   this->size = -1;
}

template<class T> 
void Vector<T>::init(int size)
{
   ASSERT(size >= 0);

   this->res = new T[size];
   this->ref_count = new int;
   *(this->ref_count) = 1;

   this->a = this->res;
   this->size = size;
}

template<class T> 
void Vector<T>::init(const Vector<T>& v)
{
   this->res = v.res;
   this->ref_count = v.ref_count;
   if (this->ref_count != NULL) {
      *(this->ref_count) = *(this->ref_count) + 1;
   }

   this->a = v.a;
   this->size = v.size;
}

template<class T> 
void Vector<T>::init(const Vector<T>& v, int offset, int size)
{
   ASSERT(!v.null());
   ASSERT(offset >= 0);
   ASSERT(size >= 0);
   ASSERT(offset <= v.size);
   ASSERT(offset + size <= v.size);

   this->res = v.res;
   this->ref_count = v.ref_count;
   *(this->ref_count) = *(this->ref_count) + 1;

   this->a = v.a + offset;
   this->size = size;
}

template<class T> 
void Vector<T>::init(const List<T>& l) 
{
   l.packInto(this);
}

template<class T> 
void Vector<T>::reinit() 
{
   purge();
   init();
}

template<class T> 
void Vector<T>::reinit(int size) 
{
   purge();
   init(size);
}

template<class T> 
void Vector<T>::reinit(const Vector<T>& v) 
{
   purge();
   init(v);
}

template<class T> 
void Vector<T>::reinit(const Vector<T>& v, int offset, int size) 
{
   purge();
   init(v,offset,size);
}

template<class T> 
void Vector<T>::reinit(const List<T>& l) 
{
   purge();
   init(l);
}

template<class T> 
void Vector<T>::purge()
{
   if (this->res != NULL) {
      this->res = NULL;
      *(this->ref_count) = *(this->ref_count) - 1;
      if (*(this->ref_count) == 0) {
         delete[] this->a;
         delete this->ref_count;
      }

      this->a = NULL;
      this->size = -1;
   }
}

template<class T> bool Vector<T>::null() const
{
   return this->res == NULL;
}

template<class T> 
bool Vector<T>::empty() const
{
   ASSERT(!this->null());
   return this->size == 0;
}

template<class T> 
T& Vector<T>::operator[](int i)
{
   ASSERT(this->a != NULL);
   ASSERT(0 <= i);
   ASSERT(i < this->size);
   return this->a[i];
}

//
// LIST. Single linked list
//
template<class T> 
struct List 
{
   struct Elem
   {
      T value;
      Elem* next;
   };
   
   Elem* getElem(Elem* e = NULL) {
      static Elem* _mem = NULL;
      if (e == NULL) {
         if (_mem == NULL) {
            return new Elem();
         } else {
            Elem* res = _mem;
            _mem = _mem->next;
            return res;
         }
      } else {
         e->next = _mem;
         _mem = e;
         return NULL;
      }
   }
   
   int size;
   Elem* head;
   Elem* tail;

   void init();

   List();
   List(List<T>& l);
   List(Vector<T>& v);

   ~List();

   void purge();
   void behead();

   bool empty();

   void append(Vector<T>& v);
   void append(List<T>& l);

   int add(T value); // returns the position of the added element

   Vector<T>* pack();
   void packInto(Vector<T>& v);

   // Disable copy and assignment;
   void operator=(const List& l);
   void operator=(const Vector<T>& v);
};


template<class T> 
List<T>::List()
{
   init();
}

template<class T> 
List<T>::List(List<T>& l)
{
   init();
   append(l);
}

template<class T> 
List<T>::List(Vector<T>& v)
{
   init();
   append(v);
}

template<class T> 
List<T>::~List()
{
   this->purge();
   delete this->head;
}

template<class T> 
void List<T>::init()
{
   this->size = 0;
   this->head = this->tail = getElem();
   this->head->next = NULL;
}

template<class T> 
void List<T>::purge()
{
   while (! empty()) { 
      behead();
   }
}

template<class T> 
void List<T>::behead()
{
   ASSERT(this->size > 0);
   
   Elem* e = this->head;
   this->head = this->head->next;
   getElem(e);
   this->size--;
}

template<class T> 
bool List<T>::empty()
{
   return size == 0;
}

template<class T> void List<T>::append(Vector<T>& v)
{
   for (int i = 0; i < v.size; i++) {
      this->add(v.a[i]);
   }
}

template<class T> 
void List<T>::append(List& l)
{
   Elem* e = l.head;
   
   while (e->next != NULL) {
      this->add(e->value);
      e = e->next;
   }
}

template<class T> 
int List<T>::add(T value)
{
   this->tail->value = value;
   this->tail->next = getElem();
   this->tail->next->next = NULL;
   this->tail = this->tail->next;
   
   return this->size++;
}

template<class T> 
Vector<T>* List<T>::pack()
{
   Vector<T>* v = new Vector<T>();

   packInto(*v);

   return v;
}


template<class T> 
void List<T>::packInto(Vector<T>& v)
{
   v.reinit(this->size);

   T* p = v.a;
   Elem* e = this->head;
   
   while (e->next != NULL) {
      ASSERT(p - v.a < size);
      *p = e->value;
      e = e->next;
      p++;
   }
}

template<class T> 
struct Queue : public List<T> 
{
   T remove();
};


template<class T> 
T Queue<T>::remove()
{
   ASSERT(this->size > 0);
   
   T value = this->head->value;
   this->behead();

   return value;
}


template<class T> 
struct Stack : public List<T> 
{
   void push(T value);
   T    peek();
   T    pop();

};

template<class T> 
void Stack<T>::push(T value)
{
   typename List<T>::Elem* e = List<T>::getElem();
   e->value = value;
   e->next = this->head;
   this->head = e;
   this->size++;
}

template<class T> T Stack<T>::peek()
{
   ASSERT(this->size > 0);
   
   return this->head->value;
}

template<class T> 
T Stack<T>::pop()
{
   ASSERT(this->size > 0);
   
   T value = peek();
   List<T>::behead();
   
   return value;
}

template<class T> 
struct Array
{
   int  size;   // the universe size
   T*   value;  // an array storing the values.
   uint  count; // the number of cells initialized so far.
   uint* index; // index[i] -- the index of the i-th initialized cell.
   uint* init_check; // reverse index of the initialized cells:
                    //    value[i] is initialized iff init_check[i] < count
                    //    and index[init_check[i]] == i

   // Standard c'tor allocates the memory of requested length.
   Array(int size);

   // Standard d'tor frees memory. 
   ~Array();
   
   // Checks if the i-th cell is initialized (set before)
   bool is_null(int i); 
   
   // Retreives the value stored in the i-th cell. Returns the
   // defualt value for T if the cell hasn't been accessed before.
   T get(int i);

   // Stores the given value to the i-th cell. Discards any value
   // previously stored in the cell.
   void set(int i, T val);

   // Gives the reference to the i-th cell. The cell is initialized to
   // the default value for T if it hasn't been accessed before. 
   T& operator[](const int i);

private: 
   // Initializes the i-th cell to the default value for T (NULL for pointers)
   // private use only. 
   void init(int i);   

   // Disable copy and assignment;
   void operator=(const Array& l);
   Array(const Array& v);
};

template<class T>
Array<T>::Array(int size)
{
   this->size = size;
   this->count = 0;
   this->value = new T[size];
   this->init_check = new uint[size];
   this->index = new uint[size];
}

template<class T> 
Array<T>::~Array()
{
   delete[] this->value;
   delete[] this->init_check;
   delete[] this->index;
}

template<class T> 
bool Array<T>::is_null(int i)
{
   ASSERT(0 <= i && i < this->size);

   return ! (this->init_check[i] < this->count && 
             this->index[this->init_check[i]] == static_cast<uint>(i));
}

template<class T> 
void Array<T>::init(int i)
{
   ASSERT(0 <= i && i < this->size);
   ASSERT(is_null(i));

   this->value[i] = NULL;
   this->init_check[i] = this->count;
   this->index[this->count] = i;
   this->count++;
}

template<class T> void Array<T>::set(int i, T val)
{
   ASSERT(0 <= i && i < this->size);

   if (is_null(i)) {
      init(i);
   }

   value[i] = val;
}

template<class T> 
T Array<T>::get(int i)
{
   ASSERT(0 <= i && i < this->size);

   if (is_null(i)) {
      init(i);
   }

   return value[i];
}

template<class T> 
T& Array<T>::operator[](const int i)
{
   ASSERT(0 <= i && i < this->size);

   if (is_null(i)) {
      init(i);
   }

   return this->value[i];
}

template<class T>
struct HArray
{
   std::tr1::unordered_map<int,T> ht;
   int size;
   HArray(int size);
   ~HArray();
   T& operator[](const int i);

   // Disable copy and assignment;
   void operator=(const HArray& l);
   HArray(const HArray& v);
};

template<class T>
HArray<T>::HArray(int size)
{
   this->size = size;
}

template<class T>
HArray<T>::~HArray()
{
}

template<class T>
T& HArray<T>::operator[](const int i)
{
   ASSERT(0 <= i && i < size);
   return ht[i];
}

#endif //RE_COLLECTIONS_H_
