//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//

#include "glushkov.h"
#include "error.h"
#include "list.h"
#include "vector.h"

/**
 * Node of a regular expression. An abstract class specifying
 * essential methods.
 */
struct Glushkov::Node
{
   /**
    * Is the subexpression nullable i.e., does empty word blong to the
    * language defined by the subexpression.
    */
   virtual bool nullable() = 0;
   /**
    * Is the given position a 'first' in the subexpression.
    */
   virtual bool first(int i) = 0;
   /**
    * Is the given position a 'last' in the subexpression.
    */
   virtual bool last(int i) = 0;
   /**
    * Does the position 'i' follows position 'j' in the subexpression.
    */
   virtual bool follows(int i, int j) = 0;
   /**
    * A virtual destructor.
    */
   virtual ~Node() {};
};
namespace {
struct Epsilon : Glushkov::Node
{
   Epsilon()
   {
   };

   ~Epsilon()
   {
   }

   bool nullable()
   {
      return true;
   };

   bool first(int i)
   {
      return false;
   };

   bool last(int i)
   {
      return false;
   };

   bool follows(int i, int j)
   {
      return false;
   };
};

struct Label : Glushkov::Node
{
   int label_id;
   int pos_id;

   Label(int label_id, int pos_id)
   {
      this->label_id = label_id;
      this->pos_id = pos_id;
   };
   
   ~Label()
   {
   }

   bool nullable()
   {
      return false;
   };

   bool first(int i)
   {
      return pos_id == i;
   };

   bool last(int i)
   {
      return pos_id == i;
   };

   bool follows(int i, int j)
   {
      return false;
   };
};


struct Star : Glushkov::Node
{
   Node* left;

   Star(Node* left)
   {
      ASSERT(left != NULL);

      this->left = left;
   };

   ~Star()
   {
      delete left;
   }

   bool nullable()
   {
      return true;
   };

   bool first(int i)
   {
      return left->first(i);
   };

   bool last(int i)
   {
      return left->last(i);
   };

   bool follows(int i, int j)
   {
      return left->follows(i,j) || (left->last(i) && left->first(j));
   };
};

struct Concat : Glushkov::Node
{
   Node* left;
   Node* right;

   Concat(Node* left, Node* right)
   {
      ASSERT(left != NULL);
      ASSERT(right != NULL);

      this->left = left;
      this->right = right;
   };

   ~Concat()
   {
      delete left;
      delete right;
   }

   bool nullable()
   {
      return left->nullable() && right->nullable();
   };

   bool first(int i)
   {
      return left->first(i) || (left->nullable() && right->first(i));
   };

   bool last(int i)
   {
      return right->last(i) || (right->nullable() && left->last(i));
   };

   bool follows(int i, int j)
   {
      return left->follows(i,j) || right->follows(i,j) || (left->last(i) && right->first(j));
   };
};

struct Union : Glushkov::Node
{
   Node* left;
   Node* right;

   Union(Node* left, Node* right)
   {
      ASSERT(left != NULL);
      ASSERT(right != NULL);

      this->left = left;
      this->right = right;
   };

   ~Union()
   {
      delete left;
      delete right;
   }

   bool nullable()
   {
      return left->nullable() || right->nullable();
   };

   bool first(int i)
   {
      return left->first(i) || right->first(i);
   };

   bool last(int i)
   {
      return right->last(i) || left->last(i);
   };

   bool follows(int i, int j)
   {
      return left->follows(i,j) || right->follows(i,j);
   };
};

static Glushkov::Node* traverse(RegExpr::Node* n, List<int>& label_id_l)
{
   switch (n->type) {
   case RegExpr::EPSILON:
      return new Epsilon();
   case RegExpr::LABEL:
      return new Label(n->label_id, label_id_l.add(n->label_id));
   case RegExpr::STAR:
      return new Star(traverse(n->left,label_id_l));
   case RegExpr::CONCAT:
      return new Concat(traverse(n->left,label_id_l),traverse(n->right,label_id_l));
   case RegExpr::UNION:
      return new Union(traverse(n->left,label_id_l),traverse(n->right,label_id_l));
   default :
      ERROR("Unmatched switch case. Should not occur!");
   }
}

}; // default namespace

Glushkov::Glushkov(RegExpr* expr)
{
   this->alphabet = new Alphabet(expr->alphabet);
   this->alphabet->freeze();
   List<int> pos_label_id_l;
   this->root = traverse(expr->root,pos_label_id_l);
   this->pos_count = pos_label_id_l.size;
   pos_label_id.reinit(pos_label_id_l);
}

Glushkov::~Glushkov()
{
   delete this->root;
   delete this->alphabet;
}

bool Glushkov::is_deterministic()
{
   // Outgoing transitions of the initial state
   for (int i = 0; i < pos_count; i++) {
      for (int j = i+1; j < pos_count; j++) {
         if (root->first(i) && root->first(j) &&
             pos_label_id[i] == pos_label_id[j]) {
            return false;
         }
      }
   }

   // Outgoing transitions of positions
   for (int k = 0; k < pos_count; k++) {
      for (int i = 0; i < pos_count; i++) {
         for (int j = i+1; j < pos_count; j++) {
            if (root->follows(k,i) && root->follows(k,j) &&
                pos_label_id[i] == pos_label_id[j]) {
               return false;
            }
         }
      }
   }
   
   return true;
}

bool Glushkov::match(Vector<char*>& symbol)
{
   Vector<int> symbol_id;

   this->alphabet->translate(symbol,symbol_id);

   return this->match(symbol_id);
}

bool Glushkov::match(Vector<int>& symbol_id)
{
   if (symbol_id.size == 0) {
      return root->nullable();
   }

   Vector<bool> curr(pos_count);
   Vector<bool> next(pos_count);
   
   curr.fill(false);
   next.fill(false);
   
   // First transition
   for (int i = 0; i < pos_count; i++) {
      curr[i] = root->first(i) 
         && pos_label_id[i] == symbol_id[0];
   }

   // Subsequent transitions
   for (int k = 1; k < symbol_id.size; k++) {
      for (int i = 0; i < pos_count; i++) {
         for (int j = 0; j < pos_count; j++) {
            next[j] = next[j] || 
               (curr[i] 
                && pos_label_id[i] == symbol_id[0]
                && root->follows(i,j));
         }
      }

      swap(curr,next);
      next.fill(false);
   }
   
   // Testing final states
   bool result = false;

   for (int i = 0; i < pos_count; i++) {
      result = result || curr[i] && root->last(i);
   }
   
   return result;
}



GlushkovRecognizer::GlushkovRecognizer(Glushkov* g) :
   Recognizer(g->alphabet)
{
   ASSERT(g != NULL);
   this->g = g;
}

bool GlushkovRecognizer::match(List<int>& symbol_id)
{
   if (symbol_id.size == 0) {
      return root->nullable();
   }

   Vector<bool> curr(pos_count);
   Vector<bool> next(pos_count);
   
   curr.fill(false);
   next.fill(false);
   
   // First transition
   for (int i = 0; i < pos_count; i++) {
      curr[i] = root->first(i) 
         && pos_label_id[i] == symbol_id[0];
   }

   // Subsequent transitions
   for (int k = 1; k < symbol_id.size; k++) {
      for (int i = 0; i < pos_count; i++) {
         for (int j = 0; j < pos_count; j++) {
            next[j] = next[j] || 
               (curr[i] 
                && pos_label_id[i] == symbol_id[0]
                && root->follows(i,j));
         }
      }

      swap(curr,next);
      next.fill(false);
   }
   
   // Testing final states
   bool result = false;

   for (int i = 0; i < pos_count; i++) {
      result = result || curr[i] && root->last(i);
   }
   
   return result;
}
