//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_LIST_H_
#define LiRE_LIST_H_

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "error.h"
#include "utils.h"
#include "logger.h"

template<class T> struct Vector;

/**
 * Array list. Basic class with extensions to a queue and a stack.
 */
template<class T>
struct List {
public:
   /**
    * Cyclic array storing the values of the list. 
    */ 
   T* a;

   /**
    * Capacity of the array. This is the allocated size of the array.
    */
   int cap;

   /**
    * Value used to quickly wrap positions in the array. This value is
    * always equal to 'cap - 1', cached for quick access. For a
    * position in the array 'i' that possibly exceeds the boundaries
    * of the array, 'i & mod' indicates the corresponding
    * index in the array. This expression evaluates to 'i % cap'
    * if 'i' is positive and 'cap - (i % cap)' when 'i' is a
    * negative number.
    */
   int mod;

   /**
    * Index the previously added element.
    */ 
   int last;

   /**
    * Number of elements in the list.
    */
   int size;

   const static int init_cap = 4;
   const static int grow = 2;

   int round(int s)
   {
      // Optimize it b/c it is EXP(LOG(min_cap)).
      int res = init_cap;
      while (res < s) {
         res *= grow;
      }
      return res;
   }

   void init(int min_cap = 0)
   {
      int cap = round(min_cap);

      this->a = new T[cap];
      this->cap = cap;
      this->mod = this->cap - 1;
      this->last = -1;
      this->size = 0;
   }

   List()
   {
      init();
   }

   List(const List<T>& l)
   {
      init(l.cap);
      append(l);
   }
   
   List(const Vector<T>& v)
   {
      init(v.size);
      append(v);
   }

   void purge()
   {
      this->last = -1;
      this->size = 0;
      trim();
   }

   void trim()
   {
   }


   static void append (T* dst, int dst_cap, int dst_mod, int& dst_last, int& dst_size,
                       const T* src, int src_cap, int src_mod, int src_last, int src_size)
   {
      ASSERT(dst != NULL && dst_size <= dst_cap);
      ASSERT(src != NULL && src_size <= src_cap);
      ASSERT(dst_cap - dst_size >= src_size);
//       printf("MOVE: \n");
//       printf("FROM: ");
//       print(src,src_cap,src_mod,src_last,src_size);
//       printf("TO  : ");
//       print(dst,dst_cap,dst_mod,dst_last,dst_size);

         
      if (src_size <= 0) {
         return;
      }

      int src_first_index     = src_last - src_size + 1;
      int src_first           = src_first_index & src_mod;
      int dst_following       = (dst_last + 1) & dst_mod;

      int max_src_block = (src_first_index < 0) ? (src_cap - src_first) : (src_size);
      int max_dst_block = (dst_following + src_size < dst_cap ) ? (src_size) : 
         (dst_cap - dst_following);

      int batch = min(max_src_block,max_dst_block);
//       printf("BATCH SIZE = %i\n",batch);
      
      memcpy(dst + dst_following, src + src_first, batch * sizeof(T));

      int new_src_size = src_size - batch;
      dst_size += batch;
      dst_last = dst_following + batch - 1;

//       printf("AFTER BATCH\n");
//       printf("FROM: ");
//       print(src,src_cap,src_mod,src_last,new_src_size);
//       printf("TO  : ");
//       print(dst,dst_cap,dst_mod,dst_last,dst_size);

      append(dst,dst_cap,dst_mod,dst_last,dst_size,
             src,src_cap,src_mod,src_last,new_src_size);
      
//       printf("RES : ");
//       print(dst,dst_cap,dst_mod,dst_last,dst_size);
   }

   void expand(int req_cap)
   {
      if (cap >= req_cap) {
         return;
      }

      int new_cap = round(req_cap);
      T* new_a = new T[new_cap];
      int new_mod = new_cap - 1;
      int new_size = 0;
      int new_last = -1;

      append(new_a,new_cap,new_mod,new_last,new_size,
             a,cap,mod,last,size);
      
      delete[] a;

      a = new_a;
      cap = new_cap;
      mod = new_mod;
      last = new_last;
      size = new_size;
   }

   bool empty()
   {
      return size == 0;
   }

   int add(T val)
   {
      if (size == cap) {
         expand(grow * cap);
      }

      last = (last + 1) & mod;
      a[last] = val;

      return size++;
   }

   void append(const List<T>& l)
   {
      expand(size+l.size);
      append(a,cap,mod,last,size,
             l.a,l.cap,l.mod,l.last,l.size);
   }

   void append(const Vector<T>& l)
   {
      expand(size+l.size);
      append(a,cap,mod,last,size,
             l.a,l.size,NULL,l.size-1,l.size);
             
   }

   void pack_into(Vector<T>&v) const
   {
      v.reinit(size);
      int v_last = 0;
      int v_size = 0;
      append(v.a,v.size,NULL,v_last,v_size,
             a,cap,mod,last,size);
      //TODO : test me.
   }

   static void print(T* a, int cap, int mod, int last, int size)
   {
//        printf("c=%i,s=%i,l=%2i,fi=%2i,f=%i ",
//               cap,size,last,last-size+1,(last-size+1)&mod);

      int first = (last-size+1)&mod;
      printf("[");
      for (int i = 0; i < cap; i++) {
         if (last == -1 && i == 0) {
            printf("{}");
         }
         if (i > 0) {
            printf(",");
         }

         if (last != -1 && size != 0 && first == i) {
            printf("{");
         }
         
         if (size != 0 && 
             ((first <= last && first <= i && i <= last) || 
              (last < first && (i <= last || first <= i)))) {
            printf("%i",(int) a[i]);
         } else {
            printf("*");
         }

         if (last != -1 && last == i) {
            if (size == 0) {
               printf("{");
            }
            printf("}");
         }

      }
      printf("]\n");
   }

   void print()
   {
      print(a,cap,mod,last,size);
   }

};

template<class T>
void reverse(const List<T>& in, List<T>& out)
{
   out.purge();

   Vector<T> in_v(in);

   for (int i = in_v.size - 1; i >= 0; i--) {
      out.add(in_v[i]);
   }
}

template<class T> 
struct Queue : List<T>
{
   Queue() : List<T>() {};
   Queue(const List<T>& l) : List<T>(l) {};
   Queue(const Vector<T>& v) : List<T>(v) {};
   T remove()
   {
      ASSERT(this->size > 0);
      int first = (this->last - this->size + 1) & this->mod;
      this->size--;
      this->trim();
      return this->a[first];
   }
   T& first() const
   {
      ASSERT(this->size > 0);
      int first = (this->last - this->size + 1) & this->mod;
      return this->a[first];
   }
};


template<class T> 
struct Stack : public List<T> 
{
   Stack() : List<T>() {};
   Stack(const List<T>& l) : List<T>(l) {};
   Stack(const Vector<T>& v) : List<T>(v) {};

   void push(T value)
   {
      this->add(value);
   }
   T peek()
   {
      ASSERT(this->size > 0);
      return this->a[this->last];
   }
   T pop()
   {
      ASSERT(this->size > 0);
      T val = this->a[this->last];
      this->last = (this->last - 1) % this->mod;
      this->size--;
      this->trim();
      return val;
   }
};

#endif // LiRE_LIST_H_
