//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//

#include <limits.h>

#include "glushkov.h"
#include "error.h"
#include "list.h"
#include "vector.h"


template<class T, void(*print_T)(T&) > 
void _print_vector(Vector<T>& v)
{
   if (v.null()) {
      printf("NULL");
   } else {
      printf("[");
      for (int i = 0; i < v.size; i++) {
         print_T(v[i]);
         if (i < v.size - 1) {
            printf(", ");
         }
      }
      printf("]\n");
   }
}


void print_int(int& i)
{
   printf("%2i",i);
}

void print_bool(bool& i)
{
   printf("%2i",i);
}

void print_vector(Vector<int>& v)
{
   _print_vector<int,print_int>(v);
}

void print_vector(Vector<bool>& v)
{
   _print_vector<bool,print_bool>(v);
}

void print_vector(Vector<Vector<bool> >& v)
{
   _print_vector<Vector<bool>,_print_vector<bool,print_bool> >(v);
}


/**
 * Node of a regular expression. An abstract class specifying
 * essential methods.
 */
struct Glushkov::Node
{
   RegExpr::NType type;

   /**
    * Is the subexpression nullable i.e., does empty word blong to the
    * language defined by the subexpression.
    */
   bool nullable;
   /**
    * Is the given position a 'first' in the subexpression.
    */
   virtual void first(Vector<bool>& l) = 0;
   /**
    * Is the given position a 'last' in the subexpression.
    */
   virtual void last(Vector<bool>& l) = 0;
   /**
    * Does the position 'i' follows position 'j' in the subexpression.
    */
   virtual void follows(Vector<Vector<bool> >& t) = 0;
   /**
    * A virtual destructor.
    */
   virtual ~Node() {};

   virtual int alternations()=0;
};
namespace {
struct Epsilon : Glushkov::Node
{
   Epsilon()
   {
      this->nullable = true;
      this->type = RegExpr::EPSILON;
   };

   ~Epsilon()
   {
   }

   void first(Vector<bool>& l)
   {
      //return false;
   };

   void last(Vector<bool>& l)
   {
      //return false;
   };
   
   void follows(Vector<Vector<bool> >& t)
   {
      //return false;
   };

   int alternations()
   {
      return 0;
   }
};

struct Label : Glushkov::Node
{
   int label_id;
   int pos_id;

   Label(int label_id, int pos_id)
   {
      this->label_id = label_id;
      this->pos_id = pos_id;
      this->nullable = false;
      this->type = RegExpr::LABEL;
   };
   
   ~Label()
   {
   }

   void first(Vector<bool>& l)
   {
      l[pos_id] = true;
      //return i == pos_id;
   };

   void last(Vector<bool>& l)
   {
      l[pos_id] = true;
      //return i == pos_id;
   };

   void follows(Vector<Vector<bool> >& t)
   {
      //return false;
   };
   
   int alternations()
   {
      return 0;
   }

};


struct Star : Glushkov::Node
{
   Node* left;

   Star(Node* left)
   {
      ASSERT(left != NULL);

      this->left = left;
      this->nullable = true;
      this->type = RegExpr::STAR;
   };

   ~Star()
   {
      delete left;
   }

   void first(Vector<bool>& l)
   {
      return left->first(l);
   };
   
   void last(Vector<bool>& l)
   {
      return left->last(l);
   };
   
   void follows(Vector<Vector<bool> >& t)
   {
      Vector<bool> f(t.size,false);
      Vector<bool> l(t.size,false);
      
      left->follows(t);

      this->left->last(l);
      this->left->first(f);
      for (int i = 0; i < t.size; i++) {
         if (l[i]) {
            for (int j = 0; j < t.size; j++) {
               if (f[j]) {
                  t[i][j] = true;
               }
            }
         }
      }

   //return left->follows(i,j) || (left->last(i) && left->first(j));
   };

   int alternations()
   {
      int a = left->alternations();
      if (a < 0) {
         a = -a;
      } 

      return a;
   }

};

struct Concat : Glushkov::Node
{
   Node* left;
   Node* right;
   
   Concat(Node* left, Node* right)
   {
      ASSERT(left != NULL);
      ASSERT(right != NULL);
      
      this->left = left;
      this->right = right;
      this->nullable = left->nullable && right->nullable;
      this->type = RegExpr::CONCAT;
   };

   ~Concat()
   {
      delete left;
      delete right;
   }

   void first(Vector<bool>& l)
   {
      left->first(l);
      if (left->nullable) {
         right->first(l);
      }
      //return left->first(i) || (left->nullable() && right->first(i));
   };

   void last(Vector<bool>& r)
   {
      right->last(r);
      if (right->nullable) {
         left->last(r);
      }
      //return right->last(i) || (right->nullable() && left->last(i));
   };

   void follows(Vector<Vector<bool> >& t)
   {
      Vector<bool> f(t.size,false);
      Vector<bool> l(t.size,false);
      left->follows(t);
      right->follows(t);
      left->last(l);
      right->first(f);
      for (int i = 0; i < l.size; i++) {
         if (l[i]) {
            for (int j = 0; j < f.size; j++) {
               if (f[j]) {
                  t[i][j] = true;
               }
            }
         }
      }
      //return left->follows(i,j) || right->follows(i,j) || (left->last(i) && right->first(j));
   };

   int alternations()
   {
      int aL = left->alternations();
      int aR = right->alternations();
      if (aL < 0) {
         aL = -aL;
      }
      if (aR < 0) {
         aR = -aR;
      }
      return max(aL,aR);
   };
};

struct Union : Glushkov::Node
{
   Node* left;
   Node* right;

   Union(Node* left, Node* right)
   {
      ASSERT(left != NULL);
      ASSERT(right != NULL);

      this->left = left;
      this->right = right;
      this->nullable = left->nullable || right->nullable;
      this->type = RegExpr::UNION;      
   };

   ~Union()
   {
      delete left;
      delete right;
   }

   void first(Vector<bool>& l)
   {
      left->first(l);
      right->first(l);
      //return left->first(i) || right->first(i);
   };

   void last(Vector<bool>& l)
   {
      right->last(l);
      left->last(l);
      //return right->last(i) || left->last(i);
   };

   void follows(Vector<Vector<bool> >& t)
   {
      left->follows(t);
      right->follows(t);
      //return left->follows(i,j) || right->follows(i,j);
   };

   int alternations()
   {
      int aL = left->alternations();
      int aR = right->alternations();
      if (aL >= 0) {
         aL = -aL - 1;
      }
      if (aR >= 0) {
         aR = -aR - 1;
      }
      return min(aL,aR);
   };
};

struct Option : Glushkov::Node
{
   Node* left;

   Option(Node* left)
   {
      ASSERT(left != NULL);

      this->left = left;
      this->nullable = true;
      this->type = RegExpr::OPTION;
   };

   ~Option()
   {
      delete left;
   }

   void first(Vector<bool>& l)
   {
      left->first(l);
   };

   void last(Vector<bool>& l)
   {
      left->last(l);
   };

   void follows(Vector<Vector<bool> >& t)
   {
      left->follows(t);
   };

   int alternations()
   {
      return left->alternations();
   }
};

struct Plus : Glushkov::Node
{
   Node* left;

   Plus(Node* left)
   {
      ASSERT(left != NULL);

      this->left = left;
      this->nullable = left->nullable;
      this->type = RegExpr::PLUS;
   };

   ~Plus()
   {
      delete left;
   }

   void first(Vector<bool>& l)
   {
      return left->first(l);
   };
   
   void last(Vector<bool>& l)
   {
      return left->last(l);
   };
   
   void follows(Vector<Vector<bool> >& t)
   {
      Vector<bool> f(t.size,false);
      Vector<bool> l(t.size,false);
      
      left->follows(t);

      this->left->last(l);
      this->left->first(f);
      for (int i = 0; i < t.size; i++) {
         if (l[i]) {
            for (int j = 0; j < t.size; j++) {
               if (f[j]) {
                  t[i][j] = true;
               }
            }
         }
      }

   //return left->follows(i,j) || (left->last(i) && left->first(j));
   };

   int alternations()
   {
      int a = left->alternations();
      if (a < 0) {
         a = -a;
      } 

      return a;
   }
};


static Glushkov::Node* traverse(RegExpr::Node* n, List<int>& label_id_l)
{
   switch (n->type) {
   case RegExpr::EPSILON:
      return new Epsilon();
   case RegExpr::LABEL:
      return new Label(n->label_id, label_id_l.add(n->label_id));
   case RegExpr::STAR:
      return new Star(traverse(n->left,label_id_l));
   case RegExpr::CONCAT:
      return new Concat(traverse(n->left,label_id_l),traverse(n->right,label_id_l));
   case RegExpr::UNION:
      return new Union(traverse(n->left,label_id_l),traverse(n->right,label_id_l));
   case RegExpr::OPTION:
      return new Option(traverse(n->left,label_id_l));
   case RegExpr::PLUS:
      return new Plus(traverse(n->left,label_id_l));
   default :
      ERROR("Unmatched switch case. Should not occur!");
   }
}

}; // default namespace

Glushkov::Glushkov(RegExpr* expr)
{
   this->alphabet = new Alphabet(expr->alphabet);
   this->alphabet->freeze();
   List<int> pos_label_id_l;
   this->root = traverse(expr->root,pos_label_id_l);
   this->pos_count = pos_label_id_l.size;
   pos_label_id.reinit(pos_label_id_l);

   this->first.reinit(this->pos_count,false);
   this->last.reinit(this->pos_count,false);
   this->follows.reinit(this->pos_count);
   for (int i = 0; i < this->follows.size; i++) {
      this->follows[i].reinit(this->pos_count,false);
   }
   this->root->first(this->first);
   this->root->last(this->last);
   this->root->follows(this->follows);
   /*
   printf("Position count: %i\n",this->pos_count);
   printf("First pos: ");
   print_vector(this->first);
   printf("\n");
   printf("Follows pos: ");
   print_vector(this->follows);
   printf("\n");
   printf("Last pos: ");
   print_vector(this->last);
   printf("\n");
   */
}

Glushkov::~Glushkov()
{
   delete this->root;
   delete this->alphabet;
}

bool Glushkov::is_deterministic()
{
   // Outgoing transitions of the initial state
   for (int i = 0; i < pos_count; i++) {
      for (int j = i+1; j < pos_count; j++) {
         if (first[i] && first[j] &&
             pos_label_id[i] == pos_label_id[j]) {
            return false;
         }
      }
   }

   // Outgoing transitions of positions
   for (int k = 0; k < pos_count; k++) {
      for (int i = 0; i < pos_count; i++) {
         for (int j = i+1; j < pos_count; j++) {
            if (follows[k][i] && follows[k][j] &&
                pos_label_id[i] == pos_label_id[j]) {
               return false;
            }
         }
      }
   }
   
   return true;
}

bool Glushkov::match(Vector<String>& symbol)
{
   Vector<int> symbol_id;

   this->alphabet->translate(symbol,symbol_id);

   return this->match(symbol_id);
}

bool Glushkov::match(Vector<int>& symbol_id)
{
//    this->print();
//    printf("START match against :");
//    print_vector(symbol_id);
//    printf("\n");

   if (symbol_id.size == 0) {
      return root->nullable;
   }

   Vector<bool> curr(pos_count);
   Vector<bool> next(pos_count);
   
   curr.fill(false);
   next.fill(false);
   

   // First transition
   for (int i = 0; i < pos_count; i++) {
      curr[i] = first[i] && pos_label_id[i] == symbol_id[0];
   }
//    printf("After 1-st transition: ");
//    print_vector(curr);
//    printf("\n");

   // Subsequent transitions
   for (int k = 1; k < symbol_id.size; k++) {
      for (int i = 0; i < pos_count; i++) {
         for (int j = 0; j < pos_count; j++) {
            next[j] = next[j] || 
               (curr[i] 
                && pos_label_id[j] == symbol_id[k]
                && follows[i][j]);
         }
      }


      swap(curr,next);
      next.fill(false);

//       printf("After %i-th transition: ",k+1);
//       print_vector(curr);
//       printf("\n");
   }
   
   // Testing final states
   bool result = false;

   for (int i = 0; i < pos_count; i++) {
      result = result || (curr[i] && last[i]);
   }
   
   return result;
}

Glushkov::Stats* Glushkov::gatherStats()
{
   Stats* stats = new Stats();
   stats->occur.reinit(alphabet->size,0);
   stats->state = pos_count;
   stats->det = is_deterministic();
   stats->sigma = alphabet->size;
   for (int i = 0; i < pos_count; i++) {
      stats->occur[pos_label_id[i]]++;
   }

   if (stats->state > 0) {
      int max = -1;
      for (int i = 0; i < stats->occur.size; i++) {
         if (max < stats->occur[i]) {
            max = stats->occur[i];
         }
      }
      stats->max_occur = max;
      
      int min = INT_MAX;
      for (int i = 0; i < stats->occur.size; i++) {
         if (min > stats->occur[i]) {
            min = stats->occur[i];
         }
      }
      stats->min_occur = min;
      
      for (int i = 0; i < stats->occur.size; i++) {
         stats->avg_occur += stats->occur[i];
      }
      stats->avg_occur /= stats->sigma;

      stats->alternat = abs(this->root->alternations());
   }
      
   stats->nullable = root->nullable;
   stats->closure = root->type == RegExpr::STAR 
      || root->type == RegExpr::PLUS;
     
   for (int i=0; i < pos_count; i++) {
      if (first[i]) {
         stats->init++;
      }
   }

   for (int i = 0; i < pos_count; i++) {
      for (int j = 0; j < pos_count; j++) {
         if (follows[i][j]) {
            stats->edge++;
         }
      }
   }

   return stats;
}


void Glushkov::print()
{
   printf("Alphabet: ");
   this->alphabet->print();
   printf("\n");
   printf("First: ");
   print_vector(this->first);
   printf("\n");
   printf("Last: ");
   print_vector(this->last);
   printf("\n");
   printf("Follows: ");
   print_vector(this->follows);
   printf("\n");
   printf("Labels: ");
   print_vector(this->pos_label_id);
   printf("\n");
   printf("Nullable: %i\n",this->root->nullable);
}

GlushkovRecognizer::GlushkovRecognizer(RegExpr* e) :
   Recognizer(e->alphabet),
   g(new Glushkov(e))
{
   ASSERT(this->g != NULL);
}

GlushkovRecognizer::~GlushkovRecognizer()
{
   delete this->g;
}

bool GlushkovRecognizer::match(Vector<int>& symbol_ids)
{
   return g->match(symbol_ids);
}

DetGlushkovRecognizer::DetGlushkovRecognizer(RegExpr* e)
   : StateTransitionRecognizer<int>(e->alphabet)
{
   this-> g = new Glushkov(e);
   //ASSERT(this->->is_deterministic());
   this->nullable = this->g->root->nullable;
   this->state_count = this->g->pos_count;

   this->init_state.reinit(this->alphabet->size,-1);
   this->delta_table.reinit(this->state_count);
   for (int i = 0; i < this->state_count; i++) {
      this->delta_table[i].reinit(this->alphabet->size,-1);
   }
   this->final.reinit(this->state_count,false);

   for (int i = 0; i < this->state_count; i++) {
      if (this->g->first[i]) {
         if (this->init_state[this->g->pos_label_id[i]] != -1) {
            ERROR("This recognizer works with deterministic expressions only.");
         }
         this->init_state[this->g->pos_label_id[i]] = i;
      }
   }

   for (int i = 0; i < this->state_count; i++) {
      for (int j = 0; j < this->state_count; j++) {
         if (this->g->follows[i][j]) {
            if (delta_table[i][this->g->pos_label_id[j]] != -1) {
               ERROR("This recognizer works with deterministic expressions only.");
            }
            delta_table[i][this->g->pos_label_id[j]] = j;
         }
      }
   }

   for (int i = 0; i < this->state_count; i++) {
      if (this->g->last[i]) {
         this->final[i] = true;
      }
   }

   
}
DetGlushkovRecognizer::~DetGlushkovRecognizer()
{
}

bool DetGlushkovRecognizer::is_valid(int q)
{
   return q >= 0;
}

int DetGlushkovRecognizer::init(int symbol_id)
{
   return init_state[symbol_id];
}

int DetGlushkovRecognizer::delta(int q, int symbol_id)
{
   return delta_table[q][symbol_id];
}

bool DetGlushkovRecognizer::accepts_empty_word()
{
   return nullable;
}
bool DetGlushkovRecognizer::accepts(int state)
{
   return final[state];
}
