//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_GLUSHKOV_H_
#define LiRE_GLUSHKOV_H_

#include "regexpr.h"
#include "vector.h"
#include "alphabet.h"
#include "recognizer.h"
#include "str.h"

/**
 * Glushkov automaton construction. 
 */
struct Glushkov
{
   /**
    * Internal representation of a node of a regular expression.
    */
   struct Node;

   struct Stats {
      int state;
      int init;
      int edge;
      bool det;
      int sigma;
      Vector<int> occur;
      int max_occur;
      double avg_occur;
      int min_occur;
      int alternat;
      bool mixed;
      bool nullable;
      bool empty;
      bool any;
      bool closure;
      char* elem_name;
      char* def;

      Stats() 
         : state(0)
         , init(0)
         , edge(0)
         , det(false)
         , sigma(0)
         , occur()
         , max_occur(0)
         , avg_occur(0.0)
         , min_occur(0)
         , alternat(0)
         , mixed(false)
         , nullable(false)
         , empty(false)
         , any(false)
         , closure(false)
         , elem_name(NULL)
         , def(NULL)
      {};
   };

   Stats* gatherStats();

   /**
    * Root node of the regular expression
    */
   Node*  root;
   
   Vector<bool> first;
   Vector<bool> last;
   Vector<Vector<bool> > follows;

   /**
    * Number of positions (non-epsilon leaves of the regular
    * expression).
    */
   int pos_count;

   /**
    * Alphabet of the regular expression. Only the elements in the
    * alphabet can be used as symbols in the regular expression.
    */
   Alphabet* alphabet;

   /**
    * Identifier of the labels of the consecutive positions in the
    * regular expression.
    */
   Vector<int> pos_label_id; 

   /**
    * Standard c'tor. Copies the structure of the regular expressions
    * and the input alphabet.
    */
   Glushkov(RegExpr* expr);

   /**
    * Standard d'tor.
    */
   ~Glushkov();

   /**
    * Checks whether the regular expression is deterministic.
    */
   bool is_deterministic();

   /**
    * Checks if the input symbol sequence belongs to the language of
    * the regular expression.
    */
   bool match(Vector<String>& symbol);

   /**
    * Checks if the input sequence of symbol identifiers (relative to
    * the 'alphabet') is recognized by the given regular expression.
    */
   bool match(Vector<int>& symbol_id);

   void print();

};

struct GlushkovRecognizer : Recognizer
{
   Glushkov* g;

   GlushkovRecognizer(RegExpr* e);
   ~GlushkovRecognizer();
   virtual bool match(Vector<int>& symbol_id);
};


struct GlushkovRecognizerFactory : RecognizerFactory
{
   Recognizer* createRecognizer(RegExpr* e)
   {
      return new GlushkovRecognizer(e);
   }
};

struct DetGlushkovRecognizer : StateTransitionRecognizer<int>
{
   Glushkov* g;
   bool nullable;
   int state_count; 
   Vector<Vector<int> > delta_table; //delta : (state_id,symbol_id) -> state_id
   Vector<int> init_state; // init_state : symbol_id -> state_id
   Vector<bool> final; // last : state_id -> {true,false}

   DetGlushkovRecognizer(RegExpr* e);
   ~DetGlushkovRecognizer();
   bool is_valid(int q);
   int init(int symbol_id);
   int delta(int q, int symbol_id);
   bool accepts_empty_word();
   bool accepts(int state);
};


struct DetGlushkovRecognizerFactory : RecognizerFactory
{
   Recognizer* createRecognizer(RegExpr* e)
   {
      return new DetGlushkovRecognizer(e);
   }
};


#endif // LiRE_GLUSHKOV_H_
