//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_LCA_H_
#define LiRE_LCA_H_

#include <stdio.h>
#include "rmq.h"

/**
 * Least Common Ancestor. A template for O(1)-time computation of a
 * least common ancestor (or nearest common ancestor) of two nodes in
 * a tree. The implementation is based on using efficient RMQ
 * implementation used on the Euler walk of the tree. The template is
 * parametrized by two classes that need to satisfy specific
 * requirements:
 * - Node representing a node of the tree. It needs to implement the 
 *   following fields:
 *   -# 'left' and 'right' pointing to the left and right 
 *      child of the node.
 *   -# 'id' storing a unique identifier among nodes in the tree 
 *      (integer value). The identifiers contain a continuous range 
 *      of numbers starting from '0'.
 * - Tree needs to implement the following fields and methods::
 *   -# 'size' indicating the total number of nodes of the tree. 
 *   -# 'get_node(int id)' returning the node with the given identifier.
 *
 * The Euler walk of the tree is a traversal of the edges of the tree
 * (induced by the standard preorder traversal of the nodes of the
 * tree). During the traversal the fields 'euler_id', 'euler_depth',
 * 'euler_open', and 'euler_close' are initialized. A node may be
 * visited more than once (if it has children) and the walk records
 * the id and the depth of any node it visits. If n is the size of the
 * tree, then the length of the walk is 2*n-1.
 */
template<class Tree,class Node>
struct LCA
{
   Tree* t;
   
   /**
    * Ids of nodes visited in the Euler walk. 'euler_id[i]' -- id of
    * the i-th visited node.
    */
   Vector<int> euler_id;
   /** 
    * Depths of the nodes visited in the Euler walk. 'euler_depth[i]'
    * -- depth of the i-th visited node.
    */ 
   Vector<int> euler_depth; 
   /**
    * Positions in the Euler walk of first visits of the
    * nodes. 'euler_open[i]' -- first visit of the node with id i.
    */
   Vector<int> euler_open; 
   /**
    * Positions in the Euler walk of last visits of the
    * nodes. 'euler_close[i]' -- last visit of the node with id i.
    */
   Vector<int> euler_close; 

   PM_RMQ* rmq;

   /**
    * Standard constructor. Initializes the structure.
    */
   LCA(Tree* t)
   {
      this->t = NULL;
      this->rmq = NULL;
      init(t);
   };

   ~LCA() 
   {
      if (rmq != NULL) {
         delete rmq;
      }
   }

   void init(Tree* t) 
   {
      ASSERT(t->size > 0);

      this->t = t;
      this->euler_id.reinit(2*t->size-1);
      this->euler_depth.reinit(2*t->size-1);
      this->euler_open.reinit(t->size);
      this->euler_close.reinit(t->size);
      
      int euler_length_chk = 0;
      traverse(this->t->root,0,euler_length_chk);

      ASSERT(euler_length_chk == 2*t->size-1);

      this->rmq = new PM_RMQ(euler_depth);
   };
   
   void traverse(Node* node, int depth, int& euler_pos)
   {
      this->euler_id[euler_pos] = node->id;
      this->euler_depth[euler_pos] = depth;
      this->euler_open[node->id] = euler_pos;
      this->euler_close[node->id] = euler_pos;
      
      euler_pos++;
      
      if (node->left != NULL) {
         traverse(node->left,depth+1,euler_pos);
         
         this->euler_id[euler_pos] = node->id;
         this->euler_depth[euler_pos] = depth;
         this->euler_close[node->id] = euler_pos;

         euler_pos++;
      }
         
      if (node->right != NULL) {
         traverse(node->right,depth+1,euler_pos);
         
         this->euler_id[euler_pos] = node->id;
         this->euler_depth[euler_pos] = depth;
         this->euler_close[node->id] = euler_pos;
         
         euler_pos++;
      } 
   }

   
   /** 
    * Finds the leas common ancestor of the given pair of nodes. If
    * one node is an ancestor of the other one or if the two nodes are
    * identical, then the node is returned.
    */
   Node* get_lca(Node* x, Node* y)
   {
      ASSERT(x != NULL);
      ASSERT(y != NULL);
      
      if (x == y) {
         return x;
      }
      
      if (x->id < y->id) {
         return t->get_node(euler_id[rmq->query(euler_open[x->id],euler_open[y->id])]);
      } else {
         return get_lca(y,x);
      }
   }


   bool checkPM()
   {
      for(int i = 1; i < euler_depth.size; i++) {
         if (abs(((long long)euler_depth[i-1])-euler_depth[i]) != 1) {
            DEBUG("Have euler_depth[i-1] = %i and euler_depth[i] = %i" 
                  " and i = %i and span = %i",
                  euler_depth[i-1],euler_depth[i],i,2*euler_depth.size-1);
            return false;
         }
      }
      
      return true;
   }

//    void print()
//    {
//       printf("Euler walk: [");
//       for (int i = 0; i < this->euler_depth.size; i++) {
//          printf("%i@%i",this->euler_id[i],this->euler_depth[i]);
//          if (i < this->euler_depth.size - 1) {
//             std::cout << ", ";
//          }
//       }
//       printf("]\n");
//    }
};
#endif // LiRE_LCA_H_
