//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

#include "logger.h"

namespace {
const char LiRE_LOG_LEVEL_NAME[6][10] = {
   "FORCE",
   "CRITICAL",
   "ERROR",
   "NOTICE",
   "INFO",
   "DEBUG"
};
}; // namespace

Logger::Logger()
{
#ifndef LiRE_DEBUG_OFF_
   this->level = DEBUG_LEVEL;
#else
   this->level = INFO_LEVEL;
#endif
}


void Logger::msg(LogLevel level, const char* fileName, int lineNo,
                 const char* format, va_list ap)
{
   if (level > this->level) {
      return;
   }

   char* message = NULL;
   char* output = NULL;

   vasprintf(&message,format,ap);
   if (!fileName) {
      output = message;
   } else if (level == FORCE_LEVEL) {
      asprintf(&output,"[%s:%i] %s",
               fileName,lineNo,
               message);
   } else {
      asprintf(&output,"[%s:%i] %s: %s",
               fileName,lineNo,
               LiRE_LOG_LEVEL_NAME[level],
               message);
   }

   fputs(output,stderr);
   fputs("\n",stderr);
   fflush(stdout);

   free(message);
   if (output != message) {
      free(output);
   }
}


void Logger::msg(LogLevel level, const char* fileName, int lineNo,
                 const char* format,...)
{
   va_list ap;
   va_start(ap,format);
   msg(level,fileName,lineNo,format,ap);
   va_end(ap);
}

void Logger::force(const char* fileName, int lineNo,
                   const char* format,...)
{
   va_list ap;
   va_start(ap,format);
   msg(FORCE_LEVEL,fileName,lineNo,format,ap);
   va_end(ap);
}

void Logger::critical(const char* fileName, int lineNo,
                      const char* format,...)
{
   va_list ap;
   va_start(ap,format);
   msg(CRITICAL_LEVEL,fileName,lineNo,format,ap);
   va_end(ap);
}

void Logger::error(const char* fileName, int lineNo,
                   const char* format,...)
{
   va_list ap;
   va_start(ap,format);
   msg(ERROR_LEVEL,fileName,lineNo,format,ap);
   va_end(ap);
}

void Logger::notice(const char* fileName, int lineNo,
                    const char* format,...)
{
   va_list ap;
   va_start(ap,format);
   msg(NOTICE_LEVEL,fileName,lineNo,format,ap);
   va_end(ap);
}

void Logger::info(const char* fileName, int lineNo,
                  const char* format,...)
{
   va_list ap;
   va_start(ap,format);
   msg(INFO_LEVEL,fileName,lineNo,format,ap);
   va_end(ap);
}

void Logger::debug(const char* fileName, int lineNo, 
                   const char* format,...)
{
   va_list ap;
   va_start(ap,format);
   msg(DEBUG_LEVEL,fileName,lineNo,format,ap);
   va_end(ap);
}


Logger* Logger::getLogger()
{  
   static Logger* _logger = new Logger();
   return _logger;
}
