//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/sax/SAXException.hpp>
#include <xercesc/parsers/SAXParser.hpp>
#include <xercesc/framework/XMLValidator.hpp>
#include <xercesc/framework/XMLErrorReporter.hpp>
#include <xercesc/validators/DTD/DTDValidator.hpp>
#include <xercesc/validators/DTD/DTDGrammar.hpp>
#include <xercesc/framework/XMLGrammarDescription.hpp>
#include <xercesc/framework/XMLDTDDescription.hpp>
#include <xercesc/validators/common/Grammar.hpp>
#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/util/OutOfMemoryException.hpp>
#include <xercesc/sax/AttributeList.hpp>
#include <xercesc/sax/SAXParseException.hpp>
#include <xercesc/sax/SAXException.hpp>
#include <xercesc/util/NameIdPool.hpp>
#include <xercesc/validators/DTD/DTDElementDecl.hpp>
#include <xercesc/framework/XMLContentModel.hpp>
#include <xercesc/validators/common/ContentSpecNode.hpp>
#if defined(XERCES_NEW_IOSTREAMS)
#include <iostream>
#include <fstream>
#else
#include <iostream.h>
#include <fstream.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <tr1/unordered_map>
#include <limits.h>

#include "validator.h"
#include "error.h"
#include "timer.h"
#include "glushkov.h"

Timer* matching = new Timer();
Timer* preprocessing = new Timer();


XERCES_CPP_NAMESPACE_USE

class TimedDTDValidator : public DTDValidator
{
public:
   TimedDTDValidator(XMLErrorReporter* const errReporter = NULL)
      : DTDValidator(errReporter)
   {
   }
   ~TimedDTDValidator(){}
   bool checkContent(XMLElementDecl* a, QName** ch, XMLSize_t n, XMLSize_t*i)
   {
      matching->start();
      bool result = DTDValidator::checkContent(a,ch,n,i);
      matching->stop();

      return result;
   }
   void validateElement(const XMLElementDecl* a)
   {
      matching->start();
      DTDValidator::validateElement(a);
      matching->stop();
   }
};

namespace {
XERCES_CPP_NAMESPACE_USE
// ---------------------------------------------------------------------------
//  This is a simple class that lets us do easy (though not terribly efficient)
//  trancoding of XMLCh data to local code page for display.
// ---------------------------------------------------------------------------
class StrX
{
public :
    // -----------------------------------------------------------------------
    //  Constructors and Destructor
    // -----------------------------------------------------------------------
    StrX(const XMLCh* const toTranscode)
    {
        // Call the private transcoding method
        fLocalForm = XMLString::transcode(toTranscode);
    }

    ~StrX()
    {
        XMLString::release(&fLocalForm);
    }

    // -----------------------------------------------------------------------
    //  Getter methods
    // -----------------------------------------------------------------------
    const char* localForm() const
    {
        return fLocalForm;
    }

   const char* str() const 
   {
      return fLocalForm;
   }

private :
    // -----------------------------------------------------------------------
    //  Private data members
    //
    //  fLocalForm
    //      This is the local code page form of the string.
    // -----------------------------------------------------------------------
    char*   fLocalForm;
};

inline XERCES_STD_QUALIFIER ostream& operator<<(XERCES_STD_QUALIFIER ostream& target, const StrX& toDump)
{
    target << toDump.localForm();
    return target;
}

class SAXCountHandlers : public HandlerBase
{
public:
   // -----------------------------------------------------------------------
   //  Constructors and Destructor
   // -----------------------------------------------------------------------
   SAXCountHandlers() :

      fAttrCount(0)
      , fCharacterCount(0)
      , fElementCount(0)
      , fSpaceCount(0)
      , fSawErrors(false)
   {
   }

   ~SAXCountHandlers()
   {
   }


   // -----------------------------------------------------------------------
   //  Getter methods
   // -----------------------------------------------------------------------
   XMLSize_t getElementCount() const
   {
      return fElementCount;
   }

   XMLSize_t getAttrCount() const
   {
      return fAttrCount;
   }

   XMLSize_t getCharacterCount() const
   {
      return fCharacterCount;
   }
   
   bool getSawErrors() const
   {
      return fSawErrors;
   }
   
   XMLSize_t getSpaceCount() const
   {
      return fSpaceCount;
   }

   // -----------------------------------------------------------------------
   //  Handlers for the SAX DocumentHandler interface
   // -----------------------------------------------------------------------
   void startElement(const XMLCh* const name, AttributeList& attributes)
   {
      fElementCount++;
      fAttrCount += attributes.getLength();
   }

   void characters(const XMLCh* const chars, const XMLSize_t length)
   {
      fCharacterCount += length;
   }

   void ignorableWhitespace(const XMLCh* const chars, const XMLSize_t length)
   {
      fSpaceCount += length;
   }

   void resetDocument()
   {
      fAttrCount = 0;
      fCharacterCount = 0;
      fElementCount = 0;
      fSpaceCount = 0;
   }

    // -----------------------------------------------------------------------
    //  Handlers for the SAX ErrorHandler interface
    // -----------------------------------------------------------------------
	void warning(const SAXParseException& e)
   {
      XERCES_STD_QUALIFIER cerr 
         << "\nWarning at file " << StrX(e.getSystemId())
         << ", line " << e.getLineNumber()
         << ", char " << e.getColumnNumber()
         << "\n  Message: " << StrX(e.getMessage()) 
         << XERCES_STD_QUALIFIER endl;
   }

   void error(const SAXParseException& e)
   {
      fSawErrors = true;
      XERCES_STD_QUALIFIER cerr 
         << "\nError at file " << StrX(e.getSystemId())
         << ", line " << e.getLineNumber()
         << ", char " << e.getColumnNumber()
         << "\n  Message: " << StrX(e.getMessage()) 
         << XERCES_STD_QUALIFIER endl;
   }

   void fatalError(const SAXParseException& e)
   {
      fSawErrors = true;
      XERCES_STD_QUALIFIER cerr 
         << "\nFatal Error at file " << StrX(e.getSystemId())
         << ", line " << e.getLineNumber()
         << ", char " << e.getColumnNumber()
         << "\n  Message: " << StrX(e.getMessage()) 
         << XERCES_STD_QUALIFIER endl;
   }

   void resetErrors()
   {
      fSawErrors = false;
   }

private:
    // -----------------------------------------------------------------------
    //  Private data members
    //
    //  fAttrCount
    //  fCharacterCount
    //  fElementCount
    //  fSpaceCount
    //      These are just counters that are run upwards based on the input
    //      from the document handlers.
    //
    //  fSawErrors
    //      This is set by the error handlers, and is queryable later to
    //      see if any errors occured.
    // -----------------------------------------------------------------------
    XMLSize_t       fAttrCount;
    XMLSize_t       fCharacterCount;
    XMLSize_t       fElementCount;
    XMLSize_t       fSpaceCount;
    bool            fSawErrors;
};
}

void test();
// ---------------------------------------------------------------------------
//  Program entry point
// ---------------------------------------------------------------------------
int main(int argC, char* argV[])
{

   //Logger::getLogger()->level = Logger::INFO_LEVEL;
   //test();
   //exit(1);
   // Check command line and extract arguments.
   if (argC < 4) {
      XERCES_STD_QUALIFIER cout << "\nUsage:\n" 
                                << "   " << argV[0] << " <file_name> <test-id> <validator>"
                                << " where <validator> can be replaced by\n" 
                                << "    0 - For Xerces DTD validator \n" 
                                << "    1 - For DetTree DTD validator \n" 
                                << "    2 - For kORE DTD validator \n"
                                << "    3 - For +-bounded DTD validator \n"
                                << "    4 - For DetGlushkov DTD validator \n"
                                << XERCES_STD_QUALIFIER endl;
      return 1;
   }
   
   const char*              validator_name[5] = { "XERCES", "DET_TREE", 
                                                  "KORE", "PLUS", "GLUSHKOV" };
   const char*              xmlFile = argV[1];
   const char*              testId = argV[2];
   int                      validator_id = atol(argV[3]);
   SAXParser::ValSchemes    valScheme = SAXParser::Val_Auto;
   bool                     doNamespaces       = false;
   bool                     doSchema           = true;
   bool                     schemaFullChecking = false;
   bool                     errorOccurred = false;
   RecognizerFactory*       factory;
   XMLValidator*            validator;
   SAXParser*               parser;
   int                      repeat_count = 3;
   
   // Initialize the XML4C2 system
   try {
      XMLPlatformUtils::Initialize();
   } catch (const XMLException& toCatch) {
      XERCES_STD_QUALIFIER cerr 
         << "Error during initialization! Message:\n"
         << StrX(toCatch.getMessage()) 
         << XERCES_STD_QUALIFIER endl;
      return 1;
   }


   for (int i = 0; i < repeat_count; i++) {

   switch (validator_id) {
   case 0 : 
      validator_id = 0;
      validator = new TimedDTDValidator();
      break;
   case 1:
      factory = new DetTreeRecognizerFactory();
      validator = new Validator(factory);
      break;
   case 2:
      factory = new DetKORERecognizerFactory();
      validator = new Validator(factory);
      break;
   case 3:
      factory = new PlusFreeRecognizerFactory();
      validator = new Validator(factory);
      break;
   case 4:
      factory = new DetGlushkovRecognizerFactory();
      validator = new Validator(factory);
      break;
   default :
      ERROR("Unknown validator id %s",argV[3]);
   }
   parser = new SAXParser(validator);
   validator->setErrorReporter(parser);

   parser->setValidationScheme(valScheme);
   parser->setDoNamespaces(doNamespaces);
   parser->setDoSchema(doSchema);
   parser->setHandleMultipleImports(true);
   parser->setValidationSchemaFullChecking(schemaFullChecking);
   
   //
   //  Create our SAX handler object and install it on the parser, as the
   //  document and error handler.
   //
   SAXCountHandlers handler;
   parser->setDocumentHandler(&handler);
   parser->setErrorHandler(&handler);
   
   //
   //  Get the starting time and kick off the parse of the indicated
   //  file. Catch any exceptions that might propogate out of it.
   //
   try {
      DEBUG("Before the parse!");
      parser->parse(xmlFile);
      DEBUG("After the parse!");
   } catch (const OutOfMemoryException&) {
      //XERCES_STD_QUALIFIER cerr << "OutOfMemoryException" << XERCES_STD_QUALIFIER endl;
      errorOccurred = true;
   } catch (const XMLException& e) {
//       XERCES_STD_QUALIFIER cerr 
//          << "\nError during parsing: '" << xmlFile << "'\n"
//          << "Exception message is:  \n"
//          << StrX(e.getMessage()) << "\n" << XERCES_STD_QUALIFIER endl;
      errorOccurred = true;
   } catch (...) {
//       XERCES_STD_QUALIFIER cerr 
//          << "\nUnexpected exception during parsing: '" << xmlFile << "'\n";
      errorOccurred = true;
   }
   
   // Print out the stats that we collected and time taken
   if (!handler.getSawErrors()) {
//       XERCES_STD_QUALIFIER cout 
//          << validator_name[validator_id] << ": " << (Timer::Elapsed() / 1000.0) << " s \n";
// //          << handler.getElementCount() << " elems, "
// //          << handler.getAttrCount() << " attrs, "
// //          << handler.getSpaceCount() << " spaces, "
// //          << handler.getCharacterCount() << " chars)" << XERCES_STD_QUALIFIER endl;
   } else {
      errorOccurred = true;
   }
   }

   if (not errorOccurred) {
      printf("UPDATE TEST SET %s_PRE = %li, %s_MATCH = %li WHERE ID = '%s';\n",
             validator_name[validator_id],
             (preprocessing->elapsed / repeat_count),
             validator_name[validator_id],
             (matching->elapsed / repeat_count),
             testId);
   }

   //
   //  Delete the parser itself.  Must be done prior to calling Terminate, below.
   //
   delete parser;
   
   // And call the termination method
   XMLPlatformUtils::Terminate();
   
      return 0;
}
