//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//

#include "recognizer.h"
#include "error.h"
#include "utils.h"

namespace {
struct EmptyWordRecognizer : Recognizer
{
   EmptyWordRecognizer() : Recognizer(new Alphabet()) {};
   ~EmptyWordRecognizer() {this->alphabet->release();};
   bool match(Vector<int>& symbol_id) {return symbol_id.empty();};
};


template<bool val> struct ConstRecognizer : Recognizer
{
   ConstRecognizer() : Recognizer(new Alphabet()) {};
   ~ConstRecognizer() {this->alphabet->release();};
   bool match(Vector<int>& symbol_id) {return val;};
};
}

Recognizer* Recognizer::EMPTY_WORD_ONLY = new EmptyWordRecognizer();
Recognizer* Recognizer::EMPTY = new ConstRecognizer<false>();
Recognizer* Recognizer::UNIVERSAL = new ConstRecognizer<true>();

Recognizer::Recognizer(Alphabet* alphabet)
{
   this->alphabet = alphabet->retain();
}

Recognizer::~Recognizer()
{
   this->alphabet->release();
}

bool Recognizer::match(Vector<String>& symbol)
{
   ASSERT(!symbol.null());
   Vector<int> symbol_id;
   this->alphabet->translate(symbol,symbol_id);
   return this->match(symbol_id);
}


ClosureRecognizer::ClosureRecognizer(Alphabet* sigma)
   : Recognizer(sigma)
{
}

ClosureRecognizer::ClosureRecognizer(Vector<String>& symbols)
   : Recognizer(new Alphabet())
{
   for (int i = 0 ; i < symbols.size; i++) {
      this->alphabet->find_id(symbols[i]);
   }
}

ClosureRecognizer::~ClosureRecognizer()
{
}

bool ClosureRecognizer::match(Vector<int>& symbol_id)
{
   ASSERT(!symbol_id.null());

   for (int i = 0; i < symbol_id.size; i++) {
      if (symbol_id[i] < 0 || symbol_id[i] > alphabet->size) {
         return false;
      }
   }

   return true;
}
