//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_RECOGNIZER_H_
#define LiRE_RECOGNIZER_H_

#include "logger.h"
#include "alphabet.h"
#include "vector.h"
#include "str.h"
#include "regexpr.h"
#include <tr1/unordered_map>


/**
 * A uniform interface for recognizing a string.
 */
struct Recognizer
{
   static Recognizer* UNIVERSAL;
   static Recognizer* EMPTY;
   static Recognizer* EMPTY_WORD_ONLY;

   Alphabet* alphabet;

   Recognizer(Alphabet* alphabet);
   virtual ~Recognizer() = 0;

   
   virtual bool match(Vector<int>& symbol_ids)=0;
   bool match(Vector<String>& symbols);

   // -----------------------------------------------------------------------
   //  Unimplemented constructors and operators
   // -----------------------------------------------------------------------
   Recognizer(const Recognizer&);
   Recognizer& operator=(const Recognizer&);
};

struct RecognizerFactory
{
   virtual Recognizer* createRecognizer(RegExpr* e)=0;
};

struct ClosureRecognizer : Recognizer
{
   ClosureRecognizer(Alphabet* sigma);
   ClosureRecognizer(Vector<String>& symbols);
   ~ClosureRecognizer();
   bool match(Vector<int>& symbol_ids);
};

struct Node;
inline void print(int i);
inline void print(Node* n);
template<class S>
struct StateTransitionRecognizer : Recognizer
{
   StateTransitionRecognizer(Alphabet* alphabet)
      : Recognizer(alphabet) 
   {};
   ~StateTransitionRecognizer() {};
   virtual bool is_valid(S q)=0;
   virtual S init(int symbol_id)=0;
   virtual S delta(S q, int symbol_id)=0;
   virtual bool accepts_empty_word()=0;
   virtual bool accepts(S state)=0;
   bool match(Vector<int>& symbol_ids)
   {

      //printf("[");
      if (symbol_ids.empty()) {
         //printf("]");
         return this->accepts_empty_word();
      } 
      
      int i = 0;
      S state = this->init(symbol_ids[i++]);
      //print(state);
      //printf(",");
      while (this->is_valid(state) && i < symbol_ids.size) {
         state = delta(state,symbol_ids[i++]);
         //print(state);
         //printf(",");
      }

      //printf("]");
      return this->is_valid(state) && accepts(state);
   }
};

inline void print(int i)
{
   printf("%i",i);
};

#endif // LiRE_RECOGNIZER_H_
