//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//

#include <stddef.h>
#include <stdio.h>

#include "regexpr.h"
#include "utils.h"
#include "error.h"
#include "regexpr_lexer.h"
#include "regexpr_parser.h"

RegExpr::Node::Node(RegExpr::NType type)
{
   this->type = type;
   this->label_id = -1;
   this->left = NULL;
   this->right = NULL;
}

RegExpr::Node::Node(RegExpr::NType type, int label_id)
{
   this->type = type;
   this->label_id = label_id;
   this->left = NULL;
   this->right = NULL;
}

RegExpr::Node::Node(RegExpr::NType type, 
                    RegExpr::Node* left, 
                    RegExpr::Node* right)
{
   this->type = type;
   this->label_id = -1;
   this->left = left;
   this->right = right;
}

RegExpr::Node::~Node()
{
   if (left != NULL) {
      delete left;
   }
   if (right != NULL) {
      delete right;
   }
}

namespace { 

typedef RegExpr::Node Node;

int size(Node* n)
{
   int left_size = (n->left == NULL)?(0):(size(n->left));
   int right_size = (n->right == NULL)?(0):(size(n->right));

   return 1 + left_size + right_size;
}

void remove_opt(Node* n) 
{
   if (n == NULL) {
      return;
   }

   remove_opt(n->left);
   remove_opt(n->right);

   if (n->type == RegExpr::OPTION) {
      n->type = RegExpr::UNION;
      n->right = new Node(RegExpr::EPSILON);
   }
}


};

RegExpr::RegExpr(Alphabet* alphabet, Node* root)
{
   ASSERT(alphabet != NULL);
   ASSERT(root != NULL);

   this->alphabet = alphabet;
   this->root = root;
}

RegExpr::~RegExpr()
{
   delete this->root;
   delete this->alphabet;
}

int RegExpr::size()
{
   return ::size(root);
}

void RegExpr::remove_option()
{
   ::remove_opt(root);
}

void Node::print()
{
   switch (this->type) {
   case RegExpr::EPSILON:
      printf("EPSILON");
      break;
   case RegExpr::LABEL:
      printf("%i",this->label_id);
      break;
   case RegExpr::STAR:
      printf("(");
      this->left->print();
      printf(")*");
      break;
   case RegExpr::CONCAT:
      printf("(");
      this->left->print();
      printf(",");
      this->right->print();
      printf(")");
      break;
   case RegExpr::UNION:
      printf("(");
      this->left->print();
      printf("|");
      this->right->print();
      printf(")");
      break;
   case RegExpr::OPTION:
      printf("(");
      this->left->print();
      printf(")?");
      break;
   case RegExpr::PLUS:
      printf("(");
      this->left->print();
      printf(")+");
      break;
   }
}

void RegExpr::print()
{
   this->root->print();
   printf(" alphabet=");
   this->alphabet->print();
   printf("\n");
}

namespace {
int aux1(RegExpr::Node* n, int size, List<int>& token)
{
   int left_size, right_size, real_size, chunk_size;
   switch (n->type) {
   case RegExpr::EPSILON :
      real_size = 0;
      break;
   case RegExpr::LABEL :
      token.add(n->label_id);
      real_size = 1;
      break;
   case RegExpr::UNION : 
      if (random_bool()) {
         real_size = aux1(n->left,size,token);
      } else {
         real_size = aux1(n->right,size,token);
      }
      break;
   case RegExpr::CONCAT :
      left_size = (size > 0) ? (random_int(size)) : (0);
      real_size = aux1(n->left,left_size,token);
      right_size = (size - real_size > 0) ? 
         (size - real_size) : (0);
      real_size += aux1(n->right,right_size,token);
      break;
   case RegExpr::STAR :
      chunk_size = (size > 0) ? (random_int(size)) : (0);
      real_size = 0;
      while (real_size < size) {
         int delta = aux1(n->left,chunk_size,token);
         if (delta == 0) {
            delta = aux1(n->left,chunk_size,token); 
            if (delta == 0) {
               delta = aux1(n->left,chunk_size,token); 
               if (delta == 0) {
                  break;
               }
            }
         }
         real_size += delta;
      }
      break;
   case RegExpr::OPTION :
      if (size > 0 && random_int(size) > 2) {
         real_size = aux1(n->left,size,token);
      } else {
         real_size = 0;
      }
      break;
   default:
      ERROR("Unhandled RegExpr::NType case");
   }
   return real_size;
}

}; // namepspace

void RegExpr::random_matching_word(Vector<int>& token)
{
   ASSERT(token.size >= 0);
   List<int> token_l;

   aux1(this->root, token.size, token_l);

   token.reinit(token_l);
}

namespace {
const uint _empty_leaf_percent = 20;
const uint _union_range  = 15;
const uint _concat_range = 60;
const uint _star_range   = 85;

RegExpr::Node* aux2(int size, int lc, Alphabet* a)
{
   ASSERT(size > 0, "Wrong size given %i",size);
   
   if (size == 1) {
      if (random_int(100) < _empty_leaf_percent) {
         return new RegExpr::Node(RegExpr::EPSILON);
      } else {
         String label = String::getCanonicalString(random_int(lc));
         int label_id = a->find_id(label);
         return new RegExpr::Node(RegExpr::LABEL,label_id);
      }
   } else if (size == 2) {
      if (random_int(100) < 50) {
         return new RegExpr::Node(RegExpr::STAR,
                                  aux2(size-1,lc,a),
                                  NULL);
      } else {
         return new RegExpr::Node(RegExpr::OPTION,
                                  aux2(size-1,lc,a),
                                  NULL);
      }
   } else { 
      uint choice = random_int(100);
      int size_split = random_int(size-2)+1;
      if (choice < _union_range) {
         return new RegExpr::Node(RegExpr::UNION,
                                  aux2(size_split,lc,a),
                                  aux2(size - size_split,lc,a));
      } else if (choice < _concat_range) {
         return new RegExpr::Node(RegExpr::CONCAT,
                                  aux2(size_split,lc,a),
                                  aux2(size - size_split,lc,a));
         
      } else if (choice < _star_range) {
         return new RegExpr::Node(RegExpr::STAR,
                                  aux2(size - 1,lc,a),
                                  NULL);
      } else {
         return new RegExpr::Node(RegExpr::OPTION,
                                  aux2(size - 1,lc,a),
                                  NULL);
      }
   }
}

}; //namespace


RegExpr* RegExpr::random_re(int size, int label_count) 
{
   ASSERT(size > 0);
   ASSERT(label_count > 0);

   Alphabet* alphabet = new Alphabet();

   return new RegExpr(alphabet, aux2(size,label_count,alphabet));
}


extern Alphabet* lire_regexpr_yy_alphabet;
extern RegExpr::Node* lire_regexpr_yy_result;
int lire_regexpr_yyparse();

RegExpr* RegExpr::parse_str(const String& str)
{
   lire_regexpr_yy_alphabet = new Alphabet();
   YY_BUFFER_STATE buf = lire_regexpr_yy_scan_string(str.str);
   if (lire_regexpr_yyparse() != 0) {
      ERROR("Parse errors");
   }
   lire_regexpr_yy_delete_buffer(buf);
   lire_regexpr_yy_alphabet->freeze();
   
   return new RegExpr(lire_regexpr_yy_alphabet,
                      lire_regexpr_yy_result);
}

RegExpr* RegExpr::parse_file(const char* file_name)
{
   FILE* file = fopen(file_name,"r");
   if (file == NULL) {
      ERROR("Could not open file '%s'",file_name);
   }

   lire_regexpr_yy_alphabet = new Alphabet();
   YY_BUFFER_STATE buf = lire_regexpr_yy_create_buffer(file,YY_BUF_SIZE);
   lire_regexpr_yy_switch_to_buffer(buf);
   if (lire_regexpr_yyparse() != 0) {
      ERROR("Parse errors");
   }
   lire_regexpr_yy_delete_buffer(buf);
   lire_regexpr_yy_alphabet->freeze();
   return new RegExpr(lire_regexpr_yy_alphabet,
                      lire_regexpr_yy_result);
}
