//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_REGEXPR_H_
#define LiRE_REGEXPR_H_

#include <stddef.h>
#include "alphabet.h"

/**
 * Regular expression.  
 *
 * Very simple and most basic tree representation intendent as the
 * output of a parser (or random generator) and the input of all our
 * algorithms. Every node has its type: epsilon, label, star,
 * concatenation, and union. A non-epsilon leaf node additionally
 * stores an identifier of a label associated with the node.
 */
struct RegExpr 
{
   /**
    * Node type.
    */
   enum NType {
      /**
       * Epsilon leaf node (no label).
       */
      EPSILON = 0, 

      /**
       * Leaf node with a label.
       */
      LABEL   = 1,

      /**
       * Kleene's star (unary node).
       */
      STAR    = 2,

      /**
       * Concatenation node (binary node).
       */
      CONCAT  = 3,

      /**
       * Union node (binary node).
       */
      UNION   = 4, 

      /**
       * Option (unary node).
       */
      OPTION  = 5,
      /**
       * Positive Klene's closure. (unary)
       */
      PLUS = 6
   };
   
   /**
    * Node of the parse tree of a regular expression. It may have
    * degree 2, 1, or 0 (leaf). A unary node has its only subtree in
    * 'left' pointer and 'right' set to NULL. In particular, if 'left'
    * is NULL, then so MUST be 'right' and then the node is a leaf.
    */
   struct Node
   {
      Node* left;
      Node* right;
      NType type;
      int   label_id;

      // Standard c'tor. Initializes all fields to default values.
      Node(NType type);
      Node(NType type, int label_id);
      Node(NType type, Node* left, Node* right);

      // Standard d'tor. Disposes recursively of all descendant nodes.
      ~Node();
      void print();
   };

   /** 
    * The alphabet of the regular expression. It should store exactly the
    * set of labels used by the regular expression
    */
   Alphabet* alphabet; 

   /**
    * The root node of the regular expression
    */
   Node* root;
   
   /**
    * Standard c'tor. 
    */
   RegExpr(Alphabet* alphabet, Node* root);

   /**
    * Standard d'ctor. Destroys the nodes and the alphabet.
    */
   ~RegExpr();

   /**
    * Calculates the size of the tree i.e., the node count of the
    * tree.
    */
   int size();

   void remove_option();

   void random_matching_word(Vector<int>& token);

   static RegExpr* random_re(int size, int label_count);

   // Parsing

   static RegExpr* parse_str(const String& str);
   static RegExpr* parse_file(const char* file_name);

   // For debugging purposes.
   void print();
};

#endif // LiRE_REGEXPR_H_
