//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//

#include "rmq.h"
#include "vector.h"
#include "error.h"
#include "utils.h"

RMQ::RMQ(Vector<int>& A) : A(A)
{
}

RMQ::~RMQ()
{
}

int RMQ::query(int i, int j)
{
   ASSERT(i < A.size, "Index %i must be within bounds (<%i)",i,A.size);
   ASSERT(j < A.size, "Index %i must be within bounds (<%i)",j,A.size);
   ASSERT(i <= j, "Index %i must precede index %i",i,j); 

   int min = A[i];
   int min_i = i;
   while (i++ < j) {
      if (min > A[i]) {
         min = A[i];
         min_i = i;
      }
   }

   return min_i;
}

NaiveRMQ::NaiveRMQ(Vector<int>& A) : RMQ(A)
{
   init_lookup();
}

NaiveRMQ::~NaiveRMQ()
{
   for (int i = 0; i < A.size; i++) {
      delete[] lookup[i];
   }

   delete[] lookup;
}

void NaiveRMQ::init_lookup()
{
   this->lookup = new int*[A.size];

   for (int i = 0; i < A.size; i++) {
      lookup[i] = new int[A.size];
   }

   // width = 0
   for (int i = 0; i < A.size; i++) {
      lookup[i][i] = i;
   }

   for (int width = 1; width < A.size; width++) {
      for (int i = 0; i < A.size - width; i++) {
         if (A[i+width] < A[lookup[i][i+width-1]]) {
            lookup[i][i+width] = i+width;
         } else {
            lookup[i][i+width] = lookup[i][i+width-1];
         }
         lookup[i+width][i] = lookup[i][i+width];
      }
   }
}

int NaiveRMQ::query(int i, int j)
{
   return lookup[i][j];
}

void  ST_RMQ::init_lookup()
{
   this->lookup = new int*[A.size];

   for (int i = 0; i < A.size; i++) {
      lookup[i] = new int[log(A.size) + 1];
   }

   for (int i = 0; i < A.size; i++) {
      lookup[i][0] = i;
   }

   for (int j = 1; j < log(A.size) + 1; j++) {
      for (int i = 0; i < A.size - exp(j-1); i++) {
         if (A[lookup[i][j-1]] <= A[lookup[i + exp(j-1)][j-1]]) {
            lookup[i][j] = lookup[i][j-1];
         } else {
            lookup[i][j] = lookup[i + exp(j-1)][j-1];
         }
      }
      for (int i = A.size - exp(j-1); i < A.size; i++) {
         lookup[i][j] = lookup[i][j-1];
      }
   }
}


ST_RMQ::ST_RMQ(Vector<int>& A) : RMQ(A)
{
   init_lookup();
}

ST_RMQ::~ST_RMQ()
{
   for (int i = 0; i < A.size; i++) {
      delete[] lookup[i];
   }

   delete[] lookup;
}

int ST_RMQ::query(int i, int j)
{
   ASSERT(i <= j);

   if (i == j) {
      return i;
   }

   int k = log(j-i);

   if (A[lookup[i][k]] <= A[lookup[j-exp(k)+1][k]]) {
      return lookup[i][k];
   } else {
      return lookup[j-exp(k)+1][k];
   }
}


// A dictionary structure that maps the contents of a +/-1 block to a
// NaiveRMQ structure for the canonical representative of this block
// (and thus returning the same query result).
struct PM_RMQ::Dict
{
   Dict* _store[3];
   Dict** branch;
   NaiveRMQ* rmq;

   Dict();
   ~Dict();
   Dict* getBranch(int delta);
   NaiveRMQ* find(Vector<int> block, int& min, int& min_pos);
};

PM_RMQ::Dict::Dict() 
{
   this->branch = _store+1;
   this->branch[-1] = NULL;
   this->branch[+1] = NULL;
   this->rmq = NULL;
}

PM_RMQ::Dict::~Dict() 
{
   if (this->branch[-1]) {
      delete this->branch[-1];
   }
   if (this->branch[+1]) {
      delete this->branch[+1];
   }
   if (this->rmq) {
      delete this->rmq;
   } 
}

PM_RMQ::Dict* PM_RMQ::Dict::getBranch(int delta) 
{
   ASSERT(delta == 1 || delta == -1);

   if (branch[delta] == NULL) {
      branch[delta] = new Dict();
   }
   
   return branch[delta];
}

NaiveRMQ* PM_RMQ::Dict::find(Vector<int> block, 
                             int& min, 
                             int& min_pos)
{
   min = block[0];
   min_pos = 0;
   Dict* node = this;      
   
   for (int i = 1; i < block.size; i++) {
      if (block[i] < min) {
         min = block[i];
         min_pos = i;
      }
      
      int delta = block[i-1] - block[i];
      
      ASSERT(abs(delta) == 1,
             " where delta is %i with block[i-1] = %i," 
             " block[i]=%i, and i = %i",delta,block[i-1],block[i],i);
      
      node = node->getBranch(delta);
   }
   
   if (node->rmq == NULL) {
      node->rmq = new NaiveRMQ(block);
   }
   
   return node->rmq;
}

PM_RMQ::PM_RMQ(Vector<int> A) : RMQ(A)
{
   this->full_block_size = max(1,log(A.size) / 2);
   int full_block_count = A.size / full_block_size; 

   int last_block_size = A.size % full_block_size;
   int B_size = full_block_count;
   if (last_block_size != 0)  {
      B_size++;
   }

   B.reinit(B_size);
   B_offset.reinit(B_size);
   root   = new Dict();
   lookup = new NaiveRMQ*[B_size];

   for (int i = 0; i < full_block_count; i++) {
      Vector<int> block(A,i*full_block_size,full_block_size);
      lookup[i] = root->find(block,B[i],B_offset[i]);
   }
   
   if (last_block_size != 0) {
      Vector<int> last_block(A,full_block_count*full_block_size,last_block_size);
      lookup[full_block_count] = 
         root->find(last_block,
                    B[full_block_count],
                    B_offset[full_block_count]);
   }
   
   B_rmq  = new ST_RMQ(B);
}

PM_RMQ::~PM_RMQ()
{
   delete B_rmq;
   delete[] lookup;
   delete root;
}


inline int PM_RMQ::block(int i)
{
   return i / full_block_size;
}

inline int PM_RMQ::offset(int i)
{
   return i % full_block_size;
}

inline int PM_RMQ::addr(int block, int offset)
{
   return block * full_block_size + offset;
}

int PM_RMQ::query(int i, int j)
{
   ASSERT(i <= j,"where i = %i and j = %i",i,j);
   int min_block, min_offset;
   int can_block, can_offset;

   if (block(i) == block(j)) {
      min_block = block(i);
      min_offset = lookup[block(i)]-> query(offset(i),offset(j));
   } else {
      min_block = block(i);
      min_offset = lookup[block(i)]->query(offset(i),full_block_size-1);

      if (block(i)+1 <= block(j)-1) {
         can_block = B_rmq->query(block(i)+1,block(j)-1);
         can_offset = B_offset[can_block];
         if (A[addr(can_block,can_offset)] < A[addr(min_block,min_offset)]) {
            min_block = can_block;
            min_offset = can_offset;
         }
      }

      can_block = block(j);
      can_offset = lookup[block(j)]->query(0,offset(j));
      if (A[addr(can_block,can_offset)] < A[addr(min_block,min_offset)]) {
         min_block = can_block;
         min_offset = can_offset;
      }
   }
   
   return min_offset + min_block * full_block_size;      
}
 
