//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_RMQ_H_
#define LiRE_RMQ_H_

#include "vector.h"

/** 
 * Range minimum queries. Given a vector of integers answer range
 * queries: for two positions of the array, find the index of the
 * minimum value stored in between the two positions. This class is
 * the most straightforward implementation for use in unit tests. More
 * efficient implementations should extend this class.
 */
struct RMQ
{
   /**
    * The vector of integers. 
    */
   Vector<int> A;
   /**
    * Standard constructor. Take the vector of integers as the
    * parameter.
    */
   RMQ(Vector<int>& A);
   
   /**
    * Standard destructor.
    */
   ~RMQ();

   /** 
    * Perform a range minimum query. Given two positions 'i' and 'j'
    * the query attempts to find a position in the array 'A' that
    * stores the minimum value in the slice of 'A' in between the
    * positions 'i' and 'j' inclusive. The positions 'i' and 'j' must
    * be within the range of the vector and 'i' must precede 'j' i.e.,
    * '0 <= i && i <= j && j < A.size'.
    */
   int query(int i, int j);
};


struct NaiveRMQ : RMQ
{
   int** lookup;

   NaiveRMQ(Vector<int>& A);
   ~NaiveRMQ();
   void init_lookup();
   
   int query(int i, int j);
};

struct ST_RMQ : public RMQ
{
   // Stores precomputed (indexes of) minimas of ranges of
   // exponentially growing length.  More precisely, lookup[i,j] =
   // query(i,min(i+2^j-1,a.size))
   int** lookup;

   ST_RMQ(Vector<int>& A);
   ~ST_RMQ();
   void init_lookup();
   int query(int i, int j);
};


// RMQ structure for arrays satisfying the +/-1 property
struct PM_RMQ : RMQ
{
   // An auxiliary dictionary structure that maps the contents of a +/-1
   // block to a NaiveRMQ structure for the canonical representant of
   // this block (and thus returning the same query result).
   struct Dict;

   // Result of splitting the array A into a sequence of continuous blocks.
   Vector<int> B;        // B[i] contains the minimum of the i-th
                         // block of A.
   Vector<int> B_offset; // B_offset[i] is the relative position of
                         // the minimum B[i] in the i-th block of A.
   int  full_block_size; // the size of full-length block; the last
                         // block may be shorter.

   ST_RMQ* B_rmq;     // RMQ structure on B.

   Dict*      root;   // The root of a dictionary mapping the contents
                      // of a full-length block to a unique NaiveRMQ
                      // structure for this block. Used only at the
                      // stage of precomputing but need to be
                      // maintained for proper resource dealocation.
   NaiveRMQ** lookup; // lookup[i] points to a RMQ structure for the
                      // i-th block of A.

   PM_RMQ(Vector<int> A);
   ~PM_RMQ();

   int block(int i);
   int offset(int i);
   int addr(int block, int offset);
   int query(int i, int j);
};

#endif // LiRE_RMQ_H_
