//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#include "skel.h"
#include "tree.h"
#include "lca.h"


//
// SKELTONS
//

typedef List<Skel::Node*> SNList;
typedef Queue<Skel::Node*> SNQueue;
typedef Stack<Skel::Node*> SNStack;

Skel::Node::Node(::Node* node) 
{
   this->id = -1;
   this->node = node;
   this->parent = NULL;
   this->left = NULL;
   this->right = NULL;
   this->star = NULL;
   this->color_rfp = NULL;
   this->witness = NULL;
   this->first_pos = NULL;
   this->next = NULL;
}

Skel::Node::~Node() {
   if (left != NULL) {
      delete left;
   }
   if (right != NULL) {
      delete right;
   }
}

namespace {
bool follows_document_order(const SNList& l)
{
   Vector<Skel::Node*> snode = l;
   bool res = true;

   for (int i = 1; i < snode.size; i++) {
      ASSERT(snode[i-1] != NULL && snode[i-1]->node != NULL);
      ASSERT(snode[i] != NULL && snode[i]->node != NULL);
      res = res && (snode[i-1]->node->id < snode[i]->node->id);
   }

   return res;
}

void merge(const SNList& in1, const SNList& in2, SNList& out)
{
   ASSERT(follows_document_order(in1));
   ASSERT(follows_document_order(in2));

   out.purge();

   SNQueue in1_copy;
   in1_copy.append(in1);
   SNQueue in2_copy;
   in2_copy.append(in2);

   while (! in1_copy.empty() && ! in2_copy.empty()) {
      if (in1_copy.first()->node->id == in2_copy.first()->node->id) {
         // careful with the duplicate.
         out.add(in1_copy.remove());
         in2_copy.remove();
      } else if (in1_copy.first()->node->id < in2_copy.first()->node->id) {
         out.add(in1_copy.remove());
      } else {
         out.add(in2_copy.remove());
      }
   }
   
   out.append(in1_copy);
   out.append(in2_copy);
}

void close_under_LCA(Tree* t, const SNList& in, SNList& out)
{
   ASSERT(follows_document_order(in));

   // Copy and reverse the input list.
   SNQueue in_rev;
   reverse(in,in_rev);

   SNList out_rev;

   SNStack reef;
   reef.add(new Skel::Node(t->root));

   while (! in_rev.empty()) {
      Skel::Node* n = in_rev.remove();
      Node* lca = t->get_lca(reef.peek()->node,n->node);
      
      while ((t->get_lca(reef.peek()->node,lca)) != reef.peek()->node) {
         out_rev.add(reef.pop());
      }

      if (lca != reef.peek()->node) {
         reef.push(new Skel::Node(lca));
      }

      if (n != reef.peek()) {
         reef.push(n);
      }
   }
   
   while (! reef.empty()) {
      out_rev.add(reef.pop());
   }

   // Empty the output list
   out.purge();
   reverse(out_rev,out);

   ASSERT(follows_document_order(out));
}

// This works only if the input node list is already closed under LCA.
void star_augment_list(Tree* t, const SNList& in, SNList& out)
{
   ASSERT(follows_document_order(in));

   SNQueue in_copy;
   in_copy.append(in);

   SNList stars;
   
   // look for the first star
   while (! in_copy.empty() && in_copy.first()->node->star == NULL) {
      in_copy.remove();
   }

   if (! in_copy.empty()) {
      stars.add(new Skel::Node(in_copy.first()->node->star));
      Node* last_star = in_copy.remove()->node->star;
      
      while (! in_copy.empty())  {
         Node* star = in_copy.remove()->node->star;
         if (star != NULL && ! star->is_ancestor(last_star)) {
            stars.add(new Skel::Node(star));
            last_star = star;
         }
      }
   }

   ASSERT(follows_document_order(stars));

   out.purge();
   merge(in,stars,out);

   ASSERT(follows_document_order(out));
}

void root_last_augment_list(Tree* t, SNList& in, SNList& out)
{
   ASSERT(follows_document_order(in));
   
   SNQueue in_copy;
   in_copy.append(in);

   SNList rls;
   
   // look for the first root_last
   while (! in_copy.empty() && in_copy.first()->node->root_last == NULL) {
      in_copy.remove();
   }

   if (! in_copy.empty()) {
      rls.add(new Skel::Node(in_copy.first()->node->root_last));
      Node* last_root_last = in_copy.remove()->node->root_last;
      
      while (! in_copy.empty())  {
         Node* root_last = in_copy.remove()->node->root_last;
         if (root_last != NULL && ! root_last->is_ancestor(last_root_last)) {
            rls.add(new Skel::Node(root_last));
            last_root_last = root_last;
         }
      }
   }


   ASSERT(follows_document_order(rls));

   out.purge();
   merge(in,rls,out);
   
   ASSERT(follows_document_order(out));
}

void get_skel_node_list(Tree* t, int color_id, SNList& out)
{
   SNList tmp_1, tmp_2, tmp_3, tmp_4;

   for (int i = 0; i < t->color_pos[color_id].size; i++) {
      tmp_1.add(new Skel::Node(t->color_pos[color_id][i]));
   }

   close_under_LCA(t,tmp_1,tmp_2);
   ASSERT(follows_document_order(tmp_2));


   tmp_1.purge();
   for (int i = 0; i < t->color_rfp[color_id].size; i++) {
      Skel::Node* sn = new Skel::Node(t->color_rfp[color_id][i]);
      sn->witness = t->color_rfp_witness[color_id][i];
      tmp_1.add(sn);
   }
   ASSERT(follows_document_order(tmp_1));

   merge(tmp_1,tmp_2,tmp_3);
   ASSERT(follows_document_order(tmp_3));

   
   star_augment_list(t,tmp_3,tmp_4);
   
   out.purge();
   
   root_last_augment_list(t,tmp_4,out);
}

void construct_skeleton(Tree* t, SNList& l, Skel* skel)
{
   SNQueue q(l);

   ASSERT(follows_document_order(q));
   ASSERT(! q.empty());
   ASSERT(q.first()->node == t->root);

   skel->node.reinit(q.size);

   int id = 0;

   skel->root = q.remove();
   skel->root->id = id;
   skel->node[id] = skel->root;
   id++;

   Skel::Node* anchor = skel->root;

   while (! q.empty()) {
      Skel::Node* next = q.remove();
      next->id = id;
      skel->node[id] = next;
      id++;

      while (! next->node->is_descendant(anchor->node)) {
         ASSERT(anchor->parent != NULL);
         anchor = anchor->parent;
      }

      if (next->node->is_left_descendant(anchor->node)) {
         ASSERT(anchor->left == NULL);
         anchor->left = next;
         anchor->left->parent = anchor;
         anchor = anchor->left;
      } else {
         ASSERT(next->node->is_right_descendant(anchor->node));
         ASSERT(anchor->right == NULL);
         anchor->right = next;
         anchor->right->parent = anchor;
         anchor = anchor->right;
      }
   }
}

void link_snodes(Skel::Node* n, Skel::Node* star, Skel::Node* color_rfp)
{
   if (n == NULL) {
      return;
   }

   if (n->node->is_star()) {
      star = n;
   }
   if (n->witness != NULL) {
      color_rfp = n;
   }

   n->star = star;
   n->color_rfp = color_rfp;
   
   if (n->node->is_pos()) {
      ASSERT(n->node->skel_leaf == NULL);
      n->node->skel_leaf = n;
   }
   
   link_snodes(n->left,star,color_rfp);
   link_snodes(n->right,star,color_rfp);
}

Node* annotate_first_pos(Skel::Node* sn)
{
   if (sn == NULL) {
      return NULL;
   }

   Node* n = sn->node;

   Node* fp_left  = annotate_first_pos(sn->left);
   Node* fp_right = annotate_first_pos(sn->right);

   switch (n->type) {
   case Node::POS :
      sn->first_pos = n;
      break;
   case Node::OPTION :
   case Node::STAR :
      sn->first_pos = fp_left;
      break;
   case Node::UNION :
      if (fp_left != NULL) {
         sn->first_pos = fp_left;
      } else {
         sn->first_pos = fp_right;
      }
      break;
   case Node::CONCAT :
      if (fp_left != NULL) {
         sn->first_pos = fp_left;
      } else if (! n->is_rfp()) {
         sn->first_pos = fp_right;
      }
      break;
   case Node::ROOT :
      sn->first_pos = fp_right;
      break;
   case Node::EPSILON : 
   case Node::UNK :
      ERROR("Wrong node type");
   }

   return sn->first_pos;
}

// FollowNodeSet
struct FNS 
{
   Skel::Node* context;
   Node* non_desc;
   Node* desc;
   bool  overflown; // concerns non-descendant only!
   FNS() {
      context = NULL;
      reset();
   }

   void reset() {
      non_desc = NULL;
      desc = NULL;
      overflown = false;
   }

   void set_context(Skel::Node* n)
   {
      ASSERT(n != NULL);
      ASSERT((context == NULL && n->node->is_root()) ||
             (n->parent != NULL && n->parent == context));

      context = n;
      Node* tmp = desc;
      desc = NULL;
      add(tmp);
   }

   void add(Node* n)
   {
      if (n == NULL) {
         return;
      }

      if (n->is_descendant(context->node)) {
         add_desc(n);
      } else {
         add_non_desc(n);
      }
   }

   void add_desc(Node* n)
   {
      if (n == NULL) {
         return;
      }
      
      ASSERT(n->is_descendant(context->node));

      if (desc != NULL && desc != n) {
         ASSERT(desc == NULL || desc == n,
                "I don't believe we will ever need to add "
                "two different descendats to the set.");
      }

      desc = n;
   }

   void add_non_desc(Node* n)
   {
      if (n == NULL) {
         return;
      }

      ASSERT(!n->is_descendant(context->node));

      if (non_desc != NULL && non_desc != n) {
         overflown = true;
      }

      non_desc = n;
   }
};

void annotate_next(Tree* t, Skel::Node* sn, FNS fns)
{
   if (sn == NULL) {
      return;
   }

   fns.set_context(sn);

   Node* n = sn->node;

   if (n->is_root_last()) {
      fns.reset();
      if (sn->parent->node != n->parent) {
         goto skip_concat;
      }
   }

   if (sn->parent != NULL && sn->parent->node->is_concat() &&
       sn == sn->parent->left && sn->parent->right != NULL) {
      fns.add(sn->parent->right->first_pos);
   }

 skip_concat:

   if (fns.overflown) {
      t->multiple_next_elements = true;
   }
   sn->next = fns.non_desc;

   if (n->is_star()) {
      fns.add(sn->first_pos);
   }

   ASSERT(n->type != Node::UNK && 
          n->type != Node::EPSILON,
          "Me doesn't know what to do with" 
          " a node of type %i (label id is %i)",
          n->type, n->color_id);
   
   annotate_next(t,sn->left,fns);
   annotate_next(t,sn->right,fns);

}

};

Skel::Skel(Tree* t, int color_id)
{
   SNList tmp;

   this->t = t;
   this->lca = NULL;
   this->color_id = color_id;
   get_skel_node_list(t,color_id,tmp);
   this->size = tmp.size;
   construct_skeleton(t,tmp,this);
   link_snodes(this->root,NULL,NULL);

}

Skel::~Skel() 
{
   if (this->root != NULL) {
      delete this->root;
   }
   if (this->lca != NULL) {
      delete this->lca;
   }
}

Skel::Node* Skel::get_lca(Skel::Node* x, Skel::Node* y)
{
   ASSERT(x != NULL);
   ASSERT(y != NULL);
   return this->lca->get_lca(x,y);
}

void Skel::annotate()
{
   annotate_first_pos(this->root);
   annotate_next(this->t,this->root,FNS());
}

void Skel::build_lca()
{
   this->lca = new LCA<Skel,Skel::Node>(this);
}
