//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#include "str.h"
#include "logger.h"

#include <string.h>
#include <stdio.h>

String String::Null;

String String::getCanonicalString(long int i)
{
   static char* _buf = new char[41];

   sprintf(_buf,"a%li",i);

   return String(_buf,true);
}

String::String()
{
   init();
}

String::String(long int ptr)
{
   ASSERT(ptr == 0);
   init();
}

String::String(const char* str)
{
   init(str);
}

String::String(char* str, bool copy) 
{
   init(str,copy);
}

String::String(const String& s)
{
   init(s);
}

String::~String()
{
   purge();
}

void String::init()
{
   this->str = NULL;
   this->ref_count = NULL;
   this->size = -1;
}

void String::init(const char* str)
{
   init(const_cast<char*>(str),true);
}

void String::init(char* str, bool copy)
{
   this->size = strlen(str);
   if (copy) {
      this->str = (char*) memcpy(new char[this->size+1],str,size+1);
   } else {
      this->str = str;
   }
   this->ref_count = new int(1);
}

void String::init(const String& s)
{
   if (s.ref_count != NULL) {
      this->str = s.str;
      this->size = s.size;
      this->ref_count = s.ref_count;
      *(this->ref_count) = *(this->ref_count) + 1;
   } else {
      init();
   }
}

void String::reinit()
{
   purge();
   init();
}

void String::reinit(const char* str)
{
   purge();
   init(str);
}

void String::reinit(char* str, bool copy)
{
   purge();
   init(str,copy);
}

void String::reinit(const String& s)
{
   if (this->ref_count != s.ref_count) {
      purge();
      init(s);
   }
}

void String::purge()
{
   if (this->ref_count != NULL) {
      *(this->ref_count) = *(this->ref_count) - 1;
      if ( *(this->ref_count) == 0) {
         //delete[] this->str;
         //delete this->ref_count;
      }
      this->str = NULL;
      this->ref_count = NULL;
      this->size = -1;
   }
}

bool String::null() const
{
   return this->str == NULL;
}

int String::len() const
{
   return this->size;
}


char& String::operator[] (int i)
{
   ASSERT(0 <= i && i <= this->size);
   return this->str[i];
}

String& String::operator=(const String& s)
{
   reinit(s);
   return *this;
}

String& String::operator+=(const String& s)
{
   ASSERT(!s.null());

   if (this->null()) {
      reinit(s);
   } else {
      char* buf = new char[this->size + s.size + 1];
      memcpy(buf,this->str,this->size);
      memcpy(buf+this->size,s.str,s.size);
      buf[this->size + s.size] = '\0';
      
      reinit(buf);
   }
    
   return *this;
}

String operator+(const String& s1, const String& s2)
{
   ASSERT(!s1.null());
   ASSERT(!s2.null());

   char* str = new char[s1.size + s2.size + 1];
   memcpy(str,s1.str,s1.size);
   memcpy(str+s1.size,s2.str,s2.size);
   str[s1.size+s2.size] = '\0';

   String res(str);
   
   return res;
}

bool operator==(const String& s1, const String& s2)
{
   return strcmp(s1.str,s2.str) == 0;
}

char* String::print() const
{
   static const int BUFFER_SIZE = 8*1024;
   static char buffer[BUFFER_SIZE];

   snprintf(buffer,BUFFER_SIZE,
            "[s=%i,ref=%i|%s]",
            this->size,
            (this->ref_count)?(*(this->ref_count)):(-1),this->str);
   return buffer;
}
