//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//

#include <stddef.h>
#include <stdio.h>
#include <iostream>

#include "utils.h"
#include "logger.h"
#include "error.h"
#include "tree.h"

typedef List<Node*> NList;

Tree::Tree(RegExpr* expr)
{
   ASSERT(expr != NULL);
   ASSERT(expr->alphabet != NULL);

   this->size = -1;
   this->node = NULL;
   this->root = NULL;   
   this->star_skeleton = NULL;
   this->root_last_skeleton = NULL;
   this->lca = NULL;

   this->multiple_next_elements = false;
   this->determinism_checked = false;
   this->deterministic = false;
   this->prepared_for_deterministic_matching = false;
   this->prepared_for_kore_matching = false;
   this->prepared_for_plus_free_matching = false;

   this->alphabet = NULL; 
   this->color_count = -1;

   //DEBUG("Before init");
   init(expr);
}

Tree::~Tree()
{
   if (this->alphabet != NULL) {
      delete this->alphabet;
   }
   if (this->node != NULL) {
      delete[] this->node;
   }
   if (this->star_skeleton) {
      delete this->star_skeleton;
   }
   if (this->root_last_skeleton) {
      delete this->root_last_skeleton;
   }
}

void Tree::init(RegExpr* expr)
{
   ASSERT(expr != NULL);
   ASSERT(expr->alphabet != NULL);

   //DEBUG("In init");
   this->alphabet = new Alphabet(expr->alphabet);
   this->alphabet->freeze();
   this->color_count = expr->alphabet->size;

   this->size = expr->size() + 1;
   node = new Node[this->size];

   copy_nodes(expr->root);
   link_nodes();
   set_interval_bounds();
   
   // Using a faster node collection gathering procedure. The new
   // procedure is approximately 1.8 times faster than the old one.
   gather_node_collections();
   // OBSOLETE: gather_collections();
}

Node* Tree::get_node(int i) 
{ 
   ASSERT(0 <= i and i < this->size); 

   return this->node + i;
}

//
// COPY.
// 

void  mark_nullable(Node* node);
Node* copy(Tree* t, RegExpr::Node* in, int depth, int& id);
void Tree::copy_nodes(RegExpr::Node* e)
{

   //DEBUG("Copying nodes");
   // Setup the virtual node.
   this->root = get_node(0);
   this->root->id = 0;
   this->root->color_id = -1;
   this->root->type = Node::ROOT;
   this->root->left = NULL;
   this->root->parent = NULL;
   
   // copy the tree structure
   int size_chk = 1;
   this->root->right = copy(this,e,1,size_chk);
   ASSERT(this->size == size_chk);

   mark_nullable(this->root);
}

Node* copy(Tree* t,           // the tree. 
           RegExpr::Node* in, // input raw node.
           int depth,         // the depth of the current node.
           int& id            // the id of the node.
           )
{
   if (in == NULL) {
      return NULL;
   }

   Node* out = t->get_node(id); 

   out->id = id++;

   out->type = (Node::NType) in->type;
   out->color_id = in->label_id;

   out->left = copy(t, in->left, depth+1, id);
   out->right = copy(t, in->right, depth+1, id);

   mark_nullable(out);
   
   return out;
}

void mark_nullable(Node* node)
{
   switch (node->type) {
   case Node::POS : 
      node->nullable = false; 
      break;
   case Node::OPTION :
   case Node::EPSILON : 
   case Node::STAR :
      node->nullable = true; 
      break;
   case Node::UNION :
      node->nullable = node->left->nullable or node->right->nullable;
      break;
   case Node::CONCAT :
      node->nullable = node->left->nullable and node->right->nullable;
      break;
   case Node::ROOT :
      node->nullable = node->right->nullable;
      break;
   case Node::UNK :
      ERROR("uninitialized type");
      break;
   }
}

//
// LINK. Set the deep pointers.
//

// Auxilriary structure to pass all links by value.
struct Links
{
   Node* parent; 
   Node* star; 
   Node* root_first; 
   Node* root_last;
   Node* rfp;
   Node* next_left;
};

void link_node(Node* node, Links links);
void Tree::link_nodes()
{
   this->root->parent = NULL;
   this->root->star = NULL;
   this->root->root_first = NULL;
   this->root->root_last = NULL;
   this->root->rfp = NULL;
   this->root->next_left = NULL;

//    Links links = { 
//       parent : this->root,
//       star : NULL,
//       root_first : this->root->right,
//       root_last : this->root->right,
//       rfp : this->root,
//       next_left : NULL
//    };

   Links links; 
   links.parent = this->root;
   links.star = NULL;
   links.root_first = this->root->right;
   links.root_last = this->root->right;
   links.rfp = this->root;
   links.next_left = NULL;

   link_node(this->root->right, links);
}

void link_node(Node* node, Links links)
{
   if (node == NULL) {
      return;
   }

   if (node->is_star()) {
      links.star = node;
   }

   if (links.parent->right == node and 
       links.parent->is_concat() and 
       not links.parent->left->nullable) {
      links.root_first = node;
      links.rfp = links.parent;
   }

   if (links.parent->left == node and 
       links.parent->is_concat() and 
       not links.parent->right->nullable) {
      links.root_last = node;
   }

   if (node->is_right_child() and
       links.parent->is_concat() and
       node->nullable) {
      links.next_left = links.parent;
   }

   node->parent = links.parent;
   node->star = links.star;
   node->root_first = links.root_first;
   node->root_last = links.root_last;
   node->rfp = links.rfp;
   node->next_left = links.next_left;

   links.parent = node;

   link_node(node->left,links);
   link_node(node->right,links);      
}

int set_interval(Node* n, int inter_begin);
void Tree::set_interval_bounds()
{
   set_interval(this->root,0);
}

int set_interval(Node* n, int inter_begin)
{
   if (n == NULL) {
      return inter_begin;
   }

   n->inter_begin = inter_begin;
   int midpoint = set_interval(n->left,inter_begin + 1);
   n->inter_end = set_interval(n->right, midpoint);

   return n->inter_end + 1;
}

//
// GATHER.
//

void Tree::gather_node_collections()
{
   // GATHERING NODE COLLECTIONS: 1st iteration counts global
   // collections, 2nd iteration gathers global collections and counts
   // local collections, 3rd iteration gathers local collections, and
   // 4th iteration gathers color-related local collections that need
   // to be gathered in a specific iteration order to ensure that the
   // collections follow the document order.

   // The counters for global collections.
   int pos_count = 0;
   int star_count = 0;
   int root_first_count = 0;
   int root_last_count = 0;
   int rfp_count = 0;

   // FIRST ITERATION: counting of nodes in global collections.
   for (int i = 0; i < size; i++) {
      Node* n = get_node(i);
      if (n->is_pos()) {
         n->pos_id = pos_count++;
      } 
      if (n->is_star()) {
         n->star_id = star_count++;
      }
      if (n->is_root_first()) {
         n->root_first_id = root_first_count++;
      } 
      if (n->is_root_last()) {
         n->root_last_id = root_last_count++;
      }
      if (n->is_rfp()) {
         n->rfp_id = rfp_count++;
      }
   }

   // Reinitializing the containers for global collections.
   this->pos.reinit(pos_count);
   this->star.reinit(star_count);
   this->root_first.reinit(root_first_count);
   this->root_last.reinit(root_last_count);
   this->rfp.reinit(rfp_count);

   // Initializing counters for local collections.
   Vector<int> root_first_pos_count(root_first_count,0);
   Vector<int> root_last_pos_count(root_last_count,0);
   Vector<int> rfp_pos_count(rfp_count,0);
   Vector<int> color_pos_count(this->color_count,0);
   Vector<int> color_rfp_count(this->color_count,0);
   Vector<int> color_rfp_witness_count(this->color_count,0);
   
   // SECOND ITERATION: Gathering the global collections and counting
   // there local ones.
   for (int i = 0; i < size; i++) {
      Node* n = get_node(i);
      if (n->is_pos()) {
         this->pos[n->pos_id] = n;
         n->root_first_pos_id = root_first_pos_count[n->root_first->root_first_id]++;
         n->root_last_pos_id = root_last_pos_count[n->root_last->root_last_id]++;
         n->rfp_pos_id = rfp_pos_count[n->rfp->rfp_id]++;
         n->color_pos_id = color_pos_count[n->color_id]++;
         color_rfp_count[n->color_id]++;
         color_rfp_witness_count[n->color_id]++;
      }
      if (n->is_star()) {
         this->star[n->star_id] = n;
      }
      if (n->is_root_first()) {
         this->root_first[n->root_first_id] = n;
      } 
      if (n->is_root_last()) {
         this->root_last[n->root_last_id] = n;
      }
      if (n->is_rfp()) {
         this->rfp[n->rfp_id] = n;
      }
   }

   // Initializing containers for local collections. 
   for (int i = 0; i < root_first.size; i++) {
      Node* rf = this->root_first[i];
      rf->root_first_pos.reinit(root_first_pos_count[i]);
   }
   for (int i = 0; i < root_last.size; i++) {
      Node* rl = this->root_last[i];
      rl->root_last_pos.reinit(root_last_pos_count[i]);
   }
   for (int i = 0; i < rfp.size; i++) {
      Node* r = this->rfp[i];
      r->rfp_pos.reinit(rfp_pos_count[i]);
   }
   color_pos.reinit(color_count);
   for (int i = 0; i < color_count; i++) {
      color_pos[i].reinit(color_pos_count[i]);
   }

   // THIRD ITERATION: Gathering the local collections.
   for (int i = 0; i < this->pos.size; i++) {
      Node* p = this->pos[i];
      p->root_first->root_first_pos[p->root_first_pos_id] = p;
      p->root_last->root_last_pos[p->root_last_pos_id] = p;
      p->rfp->rfp_pos[p->rfp_pos_id] = p;
      color_pos[p->color_id][p->color_pos_id] = p;
   }


   // Initializing node collections to be gathered int the fourth
   // iteration.
   color_first_pos.reinit(color_count,NULL);
   color_rfp.reinit(color_count);
   for (int i = 0; i < color_count; i++) {
      color_rfp[i].reinit(color_rfp_count[i]);
   }
   color_rfp_witness.reinit(color_count);
   for (int i = 0; i < color_count; i++) {
      color_rfp_witness[i].reinit(color_rfp_witness_count[i]);
   }
   // Resetting two counters.
   color_rfp_count.fill(0);
   color_rfp_witness_count.fill(0);

   // FOURTH ITERATION: Gathers order-sensitive color-arranged
   // collections. Decomposed into two loops: first for the root only
   // (and then 'color_first_pos' is gathered) and next for the
   // remaining rfp nodes.
   Node* r0 = rfp[0];
   for (int j = 0; j < r0->rfp_pos.size; j++) {
      Node* p = r0->rfp_pos[j];
      int color_id = p->color_id;
      color_rfp[color_id][color_rfp_count[color_id]++] = r0;
      color_rfp_witness[color_id][color_rfp_witness_count[color_id]++] = p;
      color_first_pos[color_id] = p;
   }
   for (int i = 1; i < rfp.size; i++) {
      Node* r = rfp[i];
      for (int j = 0; j < r->rfp_pos.size; j++) {
         Node* p = r->rfp_pos[j];
         int color_id = p->color_id;
         color_rfp[color_id][color_rfp_count[color_id]++] = r;
         color_rfp_witness[color_id][color_rfp_witness_count[color_id]++] = p;
      }
   }
}


//
// CHECK P1.
//

bool Tree::check_P1()
{

   for (int i = 0; i < color_count; i++) {
      for (int j = 1; j < color_rfp[i].size; j++) {
         if (color_rfp[i][j-1] == color_rfp[i][j]) {
            return false;
         }
      }
   }
   
   return true;
}


#define IS_BRANCH_TOP(n) (n->parent == NULL or n->parent->left != n)
#define IS_SEGMENT_TOP(n) ((n->is_root_last()) or IS_BRANCH_TOP(n))
#define IS_SEGMENT_BOTTOM(n) (n->is_epsilon() or n->is_pos() or n->left->is_root_last())
bool lire_debug_flag = false;


Node* assign_some_first_pos(Node* n)
{
   if (n == NULL) {
      return NULL;
   } else if (n->is_pos()) {
      n->some_first_pos = n;
   } else {
      Node* l = assign_some_first_pos(n->left);
      Node* r = assign_some_first_pos(n->right);

      if (l != NULL) {
         n->some_first_pos = l;
      } else if (r != NULL) {
         n->some_first_pos = r;
      } 
   }

   if (not n->is_root_first()) {
      return n->some_first_pos;
   } else {
      return NULL;
   }
}

void Tree::gather_some_first_pos()
{
   this->root->some_first_pos = assign_some_first_pos(this->root->right);
}

void assign_n_ref(Tree* t, Node* n, Node* n_ref, Node* y)
{ 
   if (n == NULL) {
      return;
   }
   
   if (n->is_pos()) {
      if (n->is_root_first()) {
         n_ref = n->parent->left;
      }
      t->n_ref[n->pos_id] = n_ref;
      ASSERT(n_ref->h[n->color_id] == NULL);
      n_ref->h[n->color_id] = n;
      return;
   } else {
      if (n->is_root_first() and not n->nullable) {
         n_ref = n->parent->left;
      }
      if (n->is_root_first() and n->nullable) {
         n_ref = y;
      }
      if (n->is_root_last() or 
          n->is_root_first() or 
          (IS_BRANCH_TOP(n) and n->nullable) or 
          (n->parent->is_union() and n->parent->right == n)) {
         y = n;
      }
   }
   
   assign_n_ref(t,n->left,n_ref,y);
   assign_n_ref(t,n->right,n_ref,y);
}

void Tree::gather_n_ref()
{
   this->n_ref.reinit(this->pos.size);

   if (this->root->right->is_pos()) {
      this->n_ref[this->root->right->pos_id] = this->root;
      ASSERT(this->root->h[this->root->right->color_id] == NULL);
      this->root->h[this->root->right->color_id] = this->root->right;
   }

   assign_n_ref(this,this->root->right->left,
                this->root,this->root->right);
   assign_n_ref(this,this->root->right->right,
                this->root,this->root->right);
}

void assign_next_top(Node* n, Node* y, Node* y_cand, bool in_seg_y_cand)
{
   if (n == NULL) {
      return;
   }

   if (n->is_epsilon()) {
      return;
   }

   if (n->is_pos() || IS_SEGMENT_TOP(n)) {
      if (n->is_root_first() and not n->nullable) {
         n->next_top = n->parent->left;
      } else {
         n->next_top = y;
      }
      if (n->is_pos()) {
         return;
      }
   } 

   if (n->is_root_first() or n->is_root_last()) {
      y = n;
      y_cand = NULL;
   } 

   if (y_cand != NULL and not n->nullable and n->is_concat() and in_seg_y_cand) {
      y = y_cand;
   }

   if (IS_BRANCH_TOP(n) and (n->nullable or n->parent->is_union())) {
      y_cand = n;
      if (not n->nullable and n->is_concat()) {
         y = y_cand;
      }
   }
   
   assign_next_top(n->left,y,y_cand,not IS_SEGMENT_TOP(n->left));
   assign_next_top(n->right,y,y_cand,false);
}
void Tree::gather_next_top()
{
   Node* n = this->root->right;

   if (n->is_epsilon()) {
      return;
   }

   if (n->is_pos() or IS_SEGMENT_TOP(n)) {
      n->next_top = this->root;
   }

   if (n->is_pos()) {
      return;
   }

   Node* y = n;
   Node* y_cand = NULL;

   if (IS_BRANCH_TOP(n) and n->nullable) {
      y_cand = n;
   }

   assign_next_top(n->left,y,y_cand,not IS_SEGMENT_TOP(n->left));
   assign_next_top(n->right,y,y_cand,false);
}

void Tree::prepare_for_plus_free_matching()
{
   ASSERT(is_deterministic());

   if (not prepared_for_plus_free_matching) {
      gather_some_first_pos();
      gather_n_ref();
      gather_next_top();
      prepared_for_plus_free_matching = true;
   }


#ifndef LiRE_DEBUG_OFF_
   for (int i = 0; i < this->pos.size; i++) {
      Node* p = this->pos[i];
      ASSERT(this->n_ref[p->pos_id] == find_n_ref(p));
   }

   for (int i = 1; i < size; i++) {
      Node* n = get_node(i);
      if (not n->is_epsilon() and (n->is_pos() or IS_SEGMENT_TOP(n))) {
         ASSERT(n->next_top == find_next_top(n));
      } else {
         ASSERT(n->next_top == NULL);
      }
   }
#endif

}

Node* Tree::find_segment_bottom(Node* y)
{
   ASSERT(y != NULL);
   while (not IS_SEGMENT_BOTTOM(y)) {
      y = y->left;
   }

   return y;
}


Node* Tree::find_highest_segment_non_nullable_concat(Node* y) 
{
   ASSERT(y != NULL);
   ASSERT(IS_BRANCH_TOP(y));

   Node* x = y;
   while (not IS_SEGMENT_BOTTOM(x) and 
          not (x->is_concat() and not x->nullable)) { 
      x = x->left;
   }

   if (x->is_concat() and not x->nullable) {
      return x;
   } else {
      return NULL;
   }
}

Node* Tree::find_n_ref(Node* p)
{
   ASSERT(p != NULL);
   ASSERT(p->is_pos());
   Node* rf = p->root_first;

   if (rf == this->root->right) {
      return this->root;
   } else if (not rf->nullable) {
      ASSERT(rf->parent != NULL and 
             rf->parent->is_concat() and
             rf->parent->left != NULL);
      return rf->parent->left;
   } else {
      Node* x = rf->parent;
      
      // Reach the top of the segment
      while (not IS_SEGMENT_TOP(x)) {
         x = x->parent;
      }
      
      Node* y = x;
      while (not y->is_root_first() and 
             not y->is_root_last() and
             not (IS_BRANCH_TOP(y) and y->nullable) and 
             not (y->parent != NULL and
                  y->parent->is_union() and 
                  y == y->parent->right)) {
         y = y->parent;
      }

      return y;
   }
}

Node* Tree::find_next_top(Node* n)
{
   ASSERT(n->is_pos() or IS_SEGMENT_TOP(n));

   if (n->parent == this->root) {
      return this->root;
   } else if (n->is_root_first() and not n->nullable) {
      return n->parent->left;
   } else {
      Node* y = n->parent;
      while (not (y->is_root_first() or //(1)
                  y->is_root_last() or  //(2)
                  (IS_BRANCH_TOP(y) and  // (3) (a)
                   (y->nullable or (y->parent != NULL and y->parent->is_union())) and //(b)
                   (find_highest_segment_non_nullable_concat(y) != NULL and 
                    find_highest_segment_non_nullable_concat(y)->is_ancestor(n))))) {
         y = y->parent;
      }

      return y;
   }
}

Node* Tree::find_some_first_pos(Node* n)
{
   ASSERT(n != NULL);
   Node* rf = n->root_first;
   for (int i = 0; i < rf->root_first_pos.size; i++) {
      if (this->first(rf->root_first_pos[i],n)) {
         return rf->root_first_pos[i];
      }
   }

   return NULL;
}

Node* Tree::find_next(Node* p, int symbol_id)
{
   ASSERT(p != NULL);
   ASSERT(symbol_id < color_count);

   if (symbol_id < 0) {
      return NULL;
   }

   Node* x = p;

   while (x != p->root_last) {
      if (x->h[symbol_id] != NULL and follows(x->h[symbol_id],p)) {
         return x->h[symbol_id];
      }
      x = x->next_top;
   }
   if (x->h[symbol_id] != NULL and follows(x->h[symbol_id],p)) {
      return x->h[symbol_id];
   }

   Node* z = x->some_first_pos;
   Node* y = (z != NULL) ? (this->n_ref[z->pos_id]) : (NULL);
   if (y != NULL and 
       y->h[symbol_id] != NULL and 
       follows(y->h[symbol_id],p)) {
      return y->h[symbol_id];
   }

   return NULL;
}

// Node* Tree::find_next(Node* p, int symbol_id)
// {
//    ASSERT(p != NULL);
//    ASSERT(symbol_id < color_count);

//    if (symbol_id < 0) {
//       return NULL;
//    }

//    Node* x = p;

//    while (x != p->root_last) {
//       if (x->h[symbol_id] != NULL and follows(x->h[symbol_id],p)) {
//          return x->h[symbol_id];
//       }
//       x = find_next_top(x);
//    }
//    if (x->h[symbol_id] != NULL and follows(x->h[symbol_id],p)) {
//       return x->h[symbol_id];
//    }

//    Node* z = find_some_first_pos(x);
//    Node* y = (z != NULL) ? (find_n_ref(z)) : (NULL);
//    if (y != NULL and 
//        y->h[symbol_id] != NULL and 
//        follows(y->h[symbol_id],p)) {
//       return y->h[symbol_id];
//    }

//    return NULL;
// }

// // Sebastian's implmeentation of P1. 
// // bool Tree::check_P1()
// // {
// //    bool* color_present = new bool[color_count];
// //    for (int i = 0; i < color_count; i++) {
// //       color_present[i] = false;
// //    }
// //    for (int i = 0; i < rfp.size; i++) {
// //       Node* r = rfp[i];
// //       for (int j = 0; j < r->rfp_pos.size; j++) {
// //          Node* p = r->rfp_pos[j];
// //          int color_id = p->color_id;
// //          if (color_present[color_id]) {
// //             return false;
// //          }
// //          color_present[color_id] = true;
// //       }
// //       for (int j = 0; j < r->rfp_pos.size; j++) {
// //          Node* p = r->rfp_pos[j];
// //          int color_id = p->color_id;
// //          color_present[color_id] = false;
// //       }
// //    }
// //    return true;
// // }

//
// Initialize LCA
//
void Tree::build_lca()
{
   ASSERT(this->lca == NULL); 

   this->lca = new LCA<Tree,Node>(this);
}

Node* Tree::get_lca(Node* x, Node* y) 
{
   ASSERT(x != NULL);
   ASSERT(y != NULL);
   
   return lca->get_lca(x,y);
}

//
// TESTING FIRST, LAST, and FOLLOWS.
//

// Test if the position x belongs to the first set of the node y.
bool Tree::first(Node* x, Node* y)
{
   ASSERT(x != NULL);
   ASSERT(y != NULL);
   ASSERT(x->is_pos());

   return y->is_ancestor(x) and y->root_first == x->root_first;
}

// Test if the position x belongs to the last set of the node y.
bool Tree::last(Node* x, Node* y)
{
   ASSERT(x != NULL);
   ASSERT(y != NULL);
   ASSERT(x->is_pos());

   return y->is_ancestor(x) and y->root_last == x->root_last;
}

// Test if the position x follows the position y.
bool Tree::follows(Node* x, Node* y)
{
   ASSERT(x != NULL);
   ASSERT(y != NULL);
   ASSERT(x->is_pos());
   ASSERT(y->is_pos());

   Node* lca = get_lca(x,y);

   if (lca->is_concat()) {
      if (last(y,lca->left) and first(x,lca->right)) {
         return true;
      }
   }
   if (lca->star != NULL) {
      if (last(y,lca->star) and first(x,lca->star)) {
         return true;
      }
   }

   return false;
}

void Tree::build_augmented_skeletons()
{
   ASSERT(this->skeleton.null());

   build_lca();
   this->skeleton.reinit(this->color_count);
   
   for (int i = 0; i < this->skeleton.size; i++) { 
      this->skeleton[i] = new Skel(this,i);
   }
}

void Tree::annotate_augmented_skeletons()
{
   for (int i = 0; i < skeleton.size; i++) {
      skeleton[i]->annotate();
   }
}

void Tree::build_star_skeleton()
{
   Vector<int> id;
   id.reinit(star.size);

   for (int i = 0; i < star.size; i++) {
      id[i] = star[i]->star_id;
   }

   this->star_skeleton = new USkel(star,id);
}

void Tree::build_root_last_skeleton()
{
   Vector<int> id;
   id.reinit(root_last.size);

   for (int i = 0; i < root_last.size; i++) {
      id[i] = root_last[i]->root_last_id;
   }

   this->root_last_skeleton = new USkel(root_last,id);
}

Node* Tree::get_highest_star(Node* x, Node* y)
{
   ASSERT(x != NULL);
   ASSERT(y != NULL);
   ASSERT(x->is_ancestor(y));

   if (y->star == NULL) {
      return NULL;
   }

   int x_star_id = (x->star) ? (x->star->star_id) : (-1);
   int y_star_id = y->star->star_id;
   
   if (x_star_id == y_star_id) {
      return NULL;
   }

   int higest_star_id = star_skeleton->getHighest(x_star_id,y_star_id);

   return star[higest_star_id];
}


Node* Tree::get_highest_root_last(Node* x, Node* y)
{
   ASSERT(x != NULL);
   ASSERT(y != NULL);
   ASSERT(x->is_ancestor(y));

   if (y->root_last == NULL) {
      return NULL;
   }

   int x_root_last_id = (x->root_last) ? (x->root_last->root_last_id) : (-1);
   int y_root_last_id = y->root_last->root_last_id;
   
   if (x_root_last_id == y_root_last_id) {
      return NULL;
   }

   int higest_root_last_id = root_last_skeleton->getHighest(x_root_last_id,y_root_last_id);

   return root_last[higest_root_last_id];
}

bool check_node(Tree* t, Skel::Node* sn)
{
   Node* n = sn->node;
   Node* n_wit = sn->witness;
   Node* n_fp = sn->first_pos;
   Node* n_next = sn->next;

   if (n_wit == NULL or n->is_root()) {
      return true;
   }

   if (n_next != NULL) {
      if (n->right != NULL and n->right->nullable) {
         return false;
      }

      Node* p = t->get_highest_root_last(n,n_wit);
      if (p == NULL) {
         p = n_wit;
      }
      if (p->next_left != NULL and n->is_ancestor(p->next_left)) {
         return false;
      }

      Node* s = t->get_highest_star(n,n_wit);
      if (s != NULL and 
          s->root_last != NULL and 
          s->root_last->is_ancestor(n)) {
         return false;
      }
   }

   if (n_fp != NULL) {
      Node* p = t->get_highest_root_last(n->left,n_fp);
      if (p == NULL) {
         p = n_fp;
      }
      if (p->next_left != NULL and n->is_ancestor(p->next_left)) {
         return false;
      }

      Node* s = t->get_highest_star(n,n_fp);
      if (s != NULL and s->root_last->is_ancestor(n->left)) {
         // It seems that this case is redundant. TO BE VERIFIED.
         return false;
      }

      if (sn->star != NULL 
          and sn->star->first_pos == n_fp 
          and n->root_last != NULL 
          and n->root_last->is_ancestor(n->star)) {
      
         if (not n->left->is_root_last()) {
            return false;
         }

         Node* y = t->get_highest_star(n,n_wit);
         if (y != NULL and y->root_last != NULL and
             y->root_last->is_ancestor(n->star)) {
            return false;
         }
         
         Node* z = t->get_highest_root_last(n,n_wit);
         if (z != NULL and z->next_left != NULL and
             n->is_ancestor(z->next_left)) {
            return false;
         }
      }
   }

   return true;

}

bool check_nodes(Tree* t, Skel::Node* n)
{
   if (n == NULL) {
      return true;
   }

   return check_node(t,n) 
      and check_nodes(t,n->left) 
      and check_nodes(t,n->right);
}

bool Tree::is_deterministic()
{
   if (not determinism_checked) {
      deterministic = check_P1();
      if (deterministic) {
         build_augmented_skeletons();
         annotate_augmented_skeletons();
         deterministic = not multiple_next_elements;
      }
      if (deterministic) {
         build_star_skeleton();
         build_root_last_skeleton();
      }
      for (int i = 0; deterministic and i < skeleton.size; i++) {
         deterministic = check_nodes(this,this->skeleton[i]->root);
      }
      determinism_checked = true;
   }

   return deterministic;
}

void Tree::prepare_for_deterministic_matching()
{
   if (prepared_for_deterministic_matching) {
      return;
   }

   ASSERT(determinism_checked and deterministic);

   for (int i = 0; i < this->skeleton.size; i++) {
      skeleton[i]->build_lca();
   }

   color_euler_pos.reinit(color_count);
   for (int i = 0; i < color_count; i++) {
      color_euler_pos[i] = new VEB<Skel::Node*>(lca->euler_id.size);
      Skel* skel = skeleton[i];
      for (int j = 0; j < skel->size; j++) {
         Skel::Node* r = skel->node[j];
         if (r->color_rfp == r) {
            color_euler_pos[i]->insert(lca->euler_open[r->node->id],r);
            color_euler_pos[i]->insert(lca->euler_close[r->node->id],r);
         }
      }
   }

   prepared_for_deterministic_matching = true;
}

void Tree::prepare_for_kore_matching()
{
   if (not determinism_checked) {
      is_deterministic(); 
   }
}

// void Tree::prepare_for_plus_free_matching()
// {
//    //TODO:: fill me out. 
// }

Skel::Node* Tree::get_nca(Node* x, int color_id)
{
   ASSERT(x != NULL);
   ASSERT(0 <= color_id and color_id < color_count);

   int x_euler_open  = lca->euler_open[x->id];
   int x_euler_close = lca->euler_close[x->id];
   int seek_left = color_euler_pos[color_id]->find_pred(x_euler_open);
   int seek_right = color_euler_pos[color_id]->find_succ(x_euler_close);

   if (seek_left == -1 or seek_right == -1) {
      return NULL;
   }

   Skel::Node* color_node_left = color_euler_pos[color_id]->get(seek_left);
   Skel::Node* color_node_right = color_euler_pos[color_id]->get(seek_right);

   Skel::Node* lca = skeleton[color_id]->get_lca(color_node_left,color_node_right);

   return lca->color_rfp;
}



TreeBasedRecognizer::TreeBasedRecognizer(Tree* t)
   : StateTransitionRecognizer<Node*>(t->alphabet)
{
   if (not t->is_deterministic()) {
      ERROR("The recognizer works only for deterministic expressions");
   }

   this->t = t;   
}

TreeBasedRecognizer::~TreeBasedRecognizer()
{
}

bool TreeBasedRecognizer::is_valid(Node* state)
{
   return state != NULL;
}

Node* TreeBasedRecognizer::init(int first_color)
{
   ASSERT(first_color < this->t->color_count);
   if (first_color < 0 ) {
      return NULL;
   }

   return t->color_first_pos[first_color];
}

bool TreeBasedRecognizer::accepts_empty_word()
{
   return this->t->root->nullable;
}

bool TreeBasedRecognizer::accepts(Node* state)
{
   ASSERT(state != NULL);

   return t->last(state,t->root->right);
}


DetTreeRecognizer::DetTreeRecognizer(RegExpr* e)
   : TreeBasedRecognizer(new Tree(e))
{
   this->t->prepare_for_deterministic_matching();
}

DetTreeRecognizer::~DetTreeRecognizer()
{
   delete this->t;
}

Node* DetTreeRecognizer::delta(Node* state, int color)
{
   ASSERT(color < this->t->color_count);
   ASSERT(state != NULL);

   if (color < 0) {
      return NULL;
   }

   Skel::Node* nca = t->get_nca(state,color);
   
   if (nca == NULL) {
      return NULL;
   } else if (nca->witness != NULL and t->follows(nca->witness,state)) {
      ASSERT(nca->witness->color_id == color);
      return nca->witness;
   } else if (nca->first_pos != NULL and t->follows(nca->first_pos,state)) {
      ASSERT(nca->first_pos->color_id == color);
      return nca->first_pos;
   } else if (nca->next != NULL and t->follows(nca->next,state)) {
      ASSERT(nca->next->color_id == color);
      return  nca->next;
   } else {
      return NULL;
   }
}

DetKORERecognizer::DetKORERecognizer(RegExpr* t)
   : TreeBasedRecognizer(new Tree(t))
{
   this->t->prepare_for_kore_matching();
}

DetKORERecognizer::~DetKORERecognizer()
{
   delete this->t;
}

Node* DetKORERecognizer::delta(Node* state, int color)
{
   ASSERT(state != NULL);

   ASSERT(color < t->color_count);
   if (color < 0) {
      return NULL;
   }

   NVector candidate = this->t->color_pos[color];

   for (int i = 0; i < candidate.size; i++) {
      if (t->follows(candidate[i],state)) {
         return candidate[i];
      }
   }

   return NULL;
}

PlusFreeRecognizer::PlusFreeRecognizer(RegExpr* e)
   : TreeBasedRecognizer(new Tree(e))
{
   this->t->prepare_for_plus_free_matching();
}
PlusFreeRecognizer::~PlusFreeRecognizer()
{
   delete this->t;
}

Node* PlusFreeRecognizer::delta(Node* n, int symbol_id)
{
   return this->t->find_next(n,symbol_id);
}

// CachedTreeRecognizer::CachedTreeRecognizer(Tree* t)
//    : TreeRecognizer(t)
// {
//    this->_cache.reinit(t->pos.size);
// }

// CachedTreeRecognizer::~CachedTreeRecognizer()
// {
// }

// Node* CachedTreeRecognizer::delta(Node* q, int symbol_id)
// {
//    ASSERT(q != NULL);
//    ASSERT(symbol_id < this->alphabet->size);
   
//    if (symbol_id < 0) {
//       return NULL;
//    }

//    if (_cache[q->pos_id].count(symbol_id) == 0) {
//       _cache[q->pos_id][symbol_id] = TreeRecognizer::delta(q,symbol_id);
//    }

//    return _cache[q->pos_id][symbol_id];
// }


// CachedDetKORERecognizer::CachedDetKORERecognizer(Tree* t)
//    : DetKORERecognizer(t)
// {
//    this->_cache.reinit(t->pos.size);
// }

// CachedDetKORERecognizer::~CachedDetKORERecognizer()
// {
// }

// Node* CachedDetKORERecognizer::delta(Node* q, int symbol_id)
// {
//    ASSERT(q != NULL);
//    ASSERT(symbol_id < this->alphabet->size);
   
//    if (symbol_id < 0) {
//       return NULL;
//    }

//    if (_cache[q->pos_id].count(symbol_id) == 0) {
//       _cache[q->pos_id][symbol_id] = DetKORERecognizer::delta(q,symbol_id);
//    }

//    return _cache[q->pos_id][symbol_id];
// }

void print(Node* n)
{
   printf("%i",(n)?(-1):(n->pos_id));
};
