//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_TREE_H_
#define LiRE_TREE_H_

#include "regexpr.h"
#include "vector.h"
#include "rmq.h"
#include "veb.h"
#include "lca.h"
#include "skel.h"
#include "uskel.h"
#include "recognizer.h"
#include <tr1/unordered_map>


struct Node;
typedef Vector<Node*> NVector;

// Tree node. A node may be used in one tree only.
struct Node
{
   // Type of a node:
   enum NType {
      UNK     = -1,
      EPSILON = 0, // Leaf node with epsilon label.
      POS     = 1, // Leaf node with a non-epsilon label.
      STAR    = 2, // Kleene's star.
      CONCAT  = 3, // Concatenation node.
      UNION   = 4, // Union node.
      OPTION  = 5, // Option node.
      ROOT    = 7  // Virtual root node.
   };

   // Basic node information 
   int   color_id;  // the id of the label of a non-epsilon leaf.
   NType type;      // the type of a node. 
   bool  nullable;  // is the node nullable.
   Node* left;      // the left child of the node.
   Node* right;     // the right child of the node .
   
   // Interval encoding for rapid ancestor checks.
   unsigned int inter_begin;  // interval beginning.
   unsigned int inter_end;    // interval end.
   
   // Deep pointers
   Node* parent;     // the parent node.
   Node* root_first; 
   Node* root_last;
   Node* star;
   Node* rfp;        // the lowest parent (concat) of a root first node
   Node* next_left;  // the lowest parent (concat) of a non-root first node
   
   // For positions only
   Skel::Node* skel_leaf; // the corresonding leaf node in a skeleton of
                          // the node's color.

   Node* some_first_pos; // Some position that is first w.r.t. the
                         // node (NULL if no such position exists)
   Node* n_ref;           // The Nref pointer (for +-free checking)
   Node* next_top;
   // Ordinal identifiers. Unique identifiers among a given class
   // of nodes. Additonnally, they are the positions in the arrays
   // storing all nodes of the same class.
   //
   // Class: all nodes; Collection: Tree::node
   int   id;
   // Class: all 'root first' nodes; Collection: Tree::root_first
   int root_first_id;
   // Class: all 'root last' nodes; Collection: Tree::root_last
   int root_last_id;
   // Class: all 'star' nodes; Collection: Tree::star
   int star_id;
   // Class: all 'position' nodes; Collection: Tree::pos
   int pos_id;
   // Class: all first 'position' nodes of a 'root first'
   // node. Collection: Node::root_first_pos
   int root_first_pos_id;
   // Class: all last 'position' nodes of a 'root last'
   // node. Collection: Node::root_last_pos
   int root_last_pos_id;
   // Class: all position nodes of the same color. Collection:
   // Tree::color_pos
   int color_pos_id;
   // Class: parents of 'root first' nodes. Collection: Tree::rfp
   int rfp_id;
   // Class: inherited positions of parent of 'root first' node.
   int rfp_pos_id;

   // Node collections.
   //
   // all positions that point with 'root first' to this node. Only
   // for 'root first' nodes.
   NVector root_first_pos;
   // all positions that point with 'root last' to this node. Only
   // for 'root last' nodes.
   NVector root_last_pos;
   // all positions that point with 'rfp' to this node. Only for
   // 'rfp' nodes.
   NVector rfp_pos;

   HArray<Node*> h;
   
   bool is_left_child();
   bool is_right_child();
   bool is_root_first();
   bool is_root_last();
   bool is_root();
   bool is_rfp();
   bool is_leaf();
   bool is_pos();
   bool is_epsilon();
   bool is_concat();
   bool is_union();
   bool is_star();
   bool is_descendant(Node* n);
   bool is_left_descendant(Node* n);
   bool is_right_descendant(Node* n);
   bool is_ancestor(Node* n);
   
   // Standard c'tor.
   Node();
   // Standard d'tor.
   ~Node();

   void print();
};

// Full-featured tree. 
struct Tree
{
   int   size;         // the size of the tree (all nodes except for virtual root).
   int   color_count;  // the number of distinct label (~colors) used in the tree.
   Node* root;         // a virtual root node on top of the original root node. 
   Alphabet* alphabet; // the alphabet of position labels (colors)

   // Collections: 
   Node*   node;       // all nodes (INCLUDING the virtual node).
   NVector pos;        // all positions.
   NVector root_first; // all 'root first' nodes.
   NVector root_last;  // all 'root last' nodes.
   NVector star;       // all 'star' nodes.
   NVector rfp;        // all parents of 'root first' nodes.
   NVector n_ref;      // all n_ref pointers of positions. 
   NVector next_top;   // all next_top pointer of all nodes. 
 
   Vector<NVector> color_pos; // color_pos[i] -- all positions of color with id i. 
   Vector<NVector> color_rfp; // color_rfp[i] -- all rfp's of color with id i.
   Vector<NVector> color_rfp_witness; // color_rfp_witness[i][j] --
                                      // the corresponding witness
                                      // position of color_rfp[i][j]
                                      // having color i
   NVector color_first_pos; // color_first_pos[color_id] -- the first
                            // position of the regular expression with
                            // the given color_id.

   Vector<Skel*> skeleton; // all augmented skeletons

   USkel* star_skeleton;
   USkel* root_last_skeleton;

   Vector<VEB<Skel::Node*>*> color_euler_pos; 

   LCA<Tree,Node>* lca;

   bool multiple_next_elements;
   bool determinism_checked;
   bool deterministic;
   bool prepared_for_deterministic_matching;
   bool prepared_for_kore_matching;
   bool prepared_for_plus_free_matching;

   void print();

   Tree(RegExpr* e);
   ~Tree();

   void init(RegExpr* e);

   // Copy the tree structure and set _Basic_ node information.
   void copy_nodes(RegExpr::Node* e);
   // Set internode pointers. 
   void link_nodes();
   // Set up interval bounds of nodes.
   void set_interval_bounds();

   void gather_node_collections();

   bool check_P1();

   void build_lca();

   // Building augmented skeletons
   void build_augmented_skeletons();
   void annotate_augmented_skeletons();

   // Building simple unranked skeletons
   void build_star_skeleton();
   void build_root_last_skeleton();

   // Checking determinism
   bool is_deterministic();

   // For +-free matching
   void gather_some_first_pos();
   void gather_n_ref();
   void gather_next_top();

   Node* find_segment_bottom(Node* y);
   Node* find_highest_segment_non_nullable_concat(Node* y);
   Node* find_n_ref(Node* p);
   Node* find_next_top(Node* n);
   Node* find_some_first_pos(Node* n);
   Node* find_next(Node* p, int symbol);

   // Parsing
   void prepare_for_deterministic_matching();
   void prepare_for_kore_matching();
   void prepare_for_plus_free_matching();
      

   // Utilities
   bool follows(Node* x, Node* y);
   bool first(Node* x, Node* y);
   bool last(Node* x, Node* y);

   Node*  get_node(int i);
   Node*  get_lca(Node* x, Node* y);
   Node*  get_highest_star(Node* x, Node* y);
   Node*  get_highest_root_last(Node* x, Node* y);
   Skel::Node* get_nca(Node* x, int color_id);
};


struct TreeBasedRecognizer : StateTransitionRecognizer<Node*>
{
   Tree* t;

   TreeBasedRecognizer(Tree* t);
   ~TreeBasedRecognizer();
   bool is_valid(Node* state);
   Node* init(int symbol_id);
   bool accepts_empty_word();
   bool accepts(Node* state);
};

struct DetTreeRecognizer : TreeBasedRecognizer
{
   DetTreeRecognizer(RegExpr* e);
   ~DetTreeRecognizer();
   Node* delta(Node* q, int symbol_id);
};

struct DetTreeRecognizerFactory : RecognizerFactory
{
   Recognizer* createRecognizer(RegExpr* e)
   {
      return new DetTreeRecognizer(e);
   }
};

struct DetKORERecognizer : TreeBasedRecognizer
{
   DetKORERecognizer(RegExpr* e);
   ~DetKORERecognizer();
   Node* delta(Node* q, int symbol_id);
};

struct DetKORERecognizerFactory : RecognizerFactory
{
   Recognizer* createRecognizer(RegExpr* e)
   {
      return new DetKORERecognizer(e);
   }
};

struct PlusFreeRecognizer : TreeBasedRecognizer
{
   PlusFreeRecognizer(RegExpr* e);
   ~PlusFreeRecognizer();
   Node* delta(Node* q, int symbol_id);
};

struct PlusFreeRecognizerFactory : RecognizerFactory
{
   Recognizer* createRecognizer(RegExpr* e)
   {
      return new PlusFreeRecognizer(e);
   }
};

// struct CachedTreeRecognizer : TreeRecognizer
// {
//    Vector<std::tr1::unordered_map<int,Node*> > _cache;
//    CachedTreeRecognizer(Tree*);
//    ~CachedTreeRecognizer();
//    Node* delta(Node* q, int symbol_id);
// };

// struct CachedDetKORERecognizer : DetKORERecognizer
// {
//    Vector<std::tr1::unordered_map<int,Node*> > _cache;
//    CachedDetKORERecognizer(Tree*);
//    ~CachedDetKORERecognizer();
//    Node* delta(Node* q, int symbol_id);
// };

#endif // LiRE_TREE_H_
