//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#include "uskel.h"
#include "tree.h"

USkel::Node::Node()
{
   this->id = -1;
   this->in_id = -1;
   this->out_id = -1;
   this->node = NULL;
   this->parent = NULL;
   this->left = NULL;
   this->right = NULL;
}

USkel::Node::~Node()
{
   if (this->left != NULL) {
      delete this->left;
   }
   if (this->right != NULL) {
      delete this->right;
   }
}

USkel::USkel(NVector& node, Vector<int>& node_id)
{
   build_uskel(node,node_id);
   this->lca = new LCA<USkel,USkel::Node>(this);
   ASSERT(this->lca->checkPM());
}

USkel::~USkel()
{
   if (this->root != NULL) {
      delete this->root;
   }
   if (this->lca != NULL) {
      delete lca;
   }
}

void USkel::build_uskel(NVector& node, Vector<int>& node_id)
{
   this->in_node.reinit(node.size);
   this->size = 2*node.size + 1;
   this->us_node.reinit(this->size);

   int id = 0;

   this->root = new USkel::Node();
   this->us_node[id] = this->root;
   this->root->id = id;
   id++;

   USkel::Node* anchor = this->root;

   for (int i = 0; i < node.size; i++) {
      ::Node* n = node[i];
      int n_id = node_id[i];

      while (anchor->parent != NULL and 
             not n->is_descendant(anchor->parent->node)) {
         anchor = roll_back(anchor,id);
      }

      anchor = add_new_node(anchor,n,n_id,id);
   }

   while (anchor->parent != NULL) {
      anchor = roll_back(anchor,id);
   }

   this->in_null = anchor;
   
}

USkel::Node* USkel::add_new_node(USkel::Node* anchor, ::Node* n, int node_id, int& usnode_id)
{
   anchor->node = n;
   anchor->out_id = node_id;

   anchor->left = new USkel::Node();
   anchor->left->parent = anchor;
   anchor->left->id = usnode_id;

   // register the new node
   this->us_node[usnode_id] = anchor->left;
   usnode_id++;

   // this will be registerd later on
   anchor->right = new USkel::Node();
   anchor->right->parent = anchor->parent;

   return anchor->left;
}

USkel::Node* USkel::roll_back(USkel::Node* anchor, int& usnode_id)
{
   ASSERT(anchor->parent != NULL);
   ASSERT(anchor->parent->right != NULL);

   anchor->in_id = anchor->parent->out_id;
   anchor->node = anchor->parent->node;
   this->in_node[anchor->in_id] = anchor;

   // register the next anchor node
   anchor->parent->right->id = usnode_id;
   this->us_node[usnode_id] = anchor->parent->right;
   usnode_id++;

   return anchor->parent->right;
}

USkel::Node* USkel::get_node(int id)
{
   return us_node[id];
}

USkel::Node* USkel::get_lca(USkel::Node* x, USkel::Node* y)
{
   ASSERT(x != NULL);
   ASSERT(y != NULL);

   return lca->get_lca(x,y);
}

int USkel::getHighest(int id_1, int id_2)
{
   ASSERT(-1 <= id_1 and id_1 < in_node.size);
   ASSERT(0 <= id_2 and id_2 < in_node.size);

   USkel::Node* n_1 = (id_1 != -1) ? (in_node[id_1]) : (in_null);
   USkel::Node* n_2 = in_node[id_2];

   ASSERT(id_1 == -1 or n_1->node->is_ancestor(n_2->node));

   USkel::Node* h = get_lca(n_1,n_2);
   return h->out_id;
}

