//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_UTILS_H_
#define LiRE_UTILS_H_

#include <stdlib.h>
#include <stdint.h>

#include "error.h"

// If the compiler is not GNU compliant
#ifndef __GNUC__
//#error "__GNUC_ undef"
#define __attribute__(x) 
#endif

// A shorter spelling of unsigned int. Very popular
typedef unsigned int uint;

// The following functions are not 'pure'

// Returns a random integer from '0' to 'bound-1' 
extern unsigned random_int(unsigned bound);

// Returns a random integer from '0' to 'bound-1' 
extern unsigned long random_long(unsigned long bound);

// Returns a random Boolean value. 
extern bool random_bool();

// Returns a canonical string asociated with the given integer.
// Essentially, it returns the representation of the input number in
// 26-base with digits ranging from 'A' through 'Z'. This function
// returns a pointer to an internal buffer and is not reentrant. 
extern char* canonical_label(long id);

// Declaring pure 'const' function for better 'subexpression' optimization.
static inline long min(long x, long y) __attribute__ ((const));
static inline long max(long x, long y) __attribute__ ((const));
static inline long sgn(long x)         __attribute__ ((const));
static inline char log8(uint8_t x)     __attribute__ ((const));
static inline char log16(uint16_t x)   __attribute__ ((const));
static inline char log32(uint32_t x)   __attribute__ ((const));
static inline char log64(uint64_t x)   __attribute__ ((const));
static inline int  log(int x)          __attribute__ ((const));
static inline int  exp(int x)          __attribute__ ((const));

// Slow simple implementation for reference and unit test comparison.
extern char log64_ref(uint64_t x);
extern uint64_t exp64_ref(int i);
extern uint64_t bit_mask_ref(int i);
extern uint64_t sqrt_ref(uint64_t i);

static inline long min(long x, long y) 
{
   return ((x)<(y))?(x):(y);
}

static inline long max(long x, long y) 
{
   return ((x)>(y))?(x):(y);
}

static inline long sgn(long x) 
{
   return (x>=0)?(+1):(-1);
}

template<class T>
static inline void swap(T& a, T& b)
{
   T tmp = a;
   a = b;
   b = tmp;
}

// Code adapted from "Bit Twiddling Hacks" by Sean Eron Anderson
// <seander@cs.stanford.edu>
// http://graphics.stanford.edu/~seander/bithacks.html
// Static lookup table containing the log of first 256 numbers.
#define _LT(n) n, n, n, n, n, n, n, n, n, n, n, n, n, n, n, n
static const char _log8_lookup[256] = {
   -1, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
   _LT(4), _LT(5), _LT(5), _LT(6), _LT(6), _LT(6), _LT(6),
   _LT(7), _LT(7), _LT(7), _LT(7), _LT(7), _LT(7), _LT(7), _LT(7)
};


static inline char log8(uint8_t x) 
{
   ASSERT(x > 0);

   return _log8_lookup[x];
}

static inline char log16(uint16_t x) 
{
   ASSERT(x > 0);

   register uint8_t t; 

   t = x >> 8;

   if (t == 0) {
      return log8(x);
   } else {
      return 8 + log8(t);
   }
}

static inline char log32(uint32_t x) {
   ASSERT(x > 0);

   register uint16_t t; 

   t = x >> 16;
   
   if (t == 0) {
      return log16(x);
   } else {
      return 16 + log16(t);
   }
}

static inline char log64(uint64_t x) {
   ASSERT(x > 0);

   register uint32_t t; 

   t = x >> 32;

   if (t == 0) {
      return log32((uint32_t)x);
   } else {
      return 32 + log32(t);
   }
}

static inline int log(int x)
{
   ASSERT(x > 0," and the value is %i",x);
   return log32(x);
}

static inline int exp(int x)
{
   ASSERT(x < 31);
   return 1 << x;
}

extern size_t str_len(const char* str);
extern const char* str_dup(const char* str);


// These two are to be moved to the printing package.
extern void indent(int id);
extern void latex_indent(int id);

#endif // LiRE_UTILS_H_
