#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/sax/SAXException.hpp>
#include <xercesc/parsers/SAXParser.hpp>
#include <xercesc/framework/XMLValidator.hpp>
#include <xercesc/framework/XMLErrorReporter.hpp>
#include <xercesc/validators/DTD/DTDValidator.hpp>
#include <xercesc/validators/DTD/DTDGrammar.hpp>
#include <xercesc/framework/XMLGrammarDescription.hpp>
#include <xercesc/framework/XMLDTDDescription.hpp>
#include <xercesc/validators/common/Grammar.hpp>
#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/util/OutOfMemoryException.hpp>
#include <xercesc/sax/AttributeList.hpp>
#include <xercesc/sax/SAXParseException.hpp>
#include <xercesc/sax/SAXException.hpp>
#include <xercesc/util/NameIdPool.hpp>
#include <xercesc/validators/DTD/DTDElementDecl.hpp>
#include <xercesc/framework/XMLContentModel.hpp>
#include <xercesc/validators/common/ContentSpecNode.hpp>
#if defined(XERCES_NEW_IOSTREAMS)
#include <iostream>
#include <fstream>
#else
#include <iostream.h>
#include <fstream.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <tr1/unordered_map>
#include <limits.h>

#include "validator.h"
#include "alphabet.h"
#include "regexpr.h"
#include "tree.h"
#include "error.h"
#include "glushkov.h"
#include "timer.h"

extern Timer* matching;
extern Timer* preprocessing;

XERCES_CPP_NAMESPACE_USE

Validator::Validator(RecognizerFactory* f, XMLErrorReporter* const errReporter) :
   XMLValidator(errReporter)
{
   this->f = f;
   this->grammar = NULL;
}

Validator::~Validator()
{
}

String& Validator::getSymbol(const XMLCh* element_name) 
{
   if (symbol[element_name].null()) {
      String s(XMLString::transcode(element_name));
      symbol[element_name] = s;
   }
      
   return symbol[element_name];
}

RegExpr::Node* Validator::translate(const ContentSpecNode* spec, Alphabet* alphabet)
{
   const XMLCh* element_name;
   RegExpr::Node* child = NULL;   

   switch (spec->getType()) {
   case ContentSpecNode::Leaf :
      element_name = spec->getElement()->getRawName();
      if (XMLString::stringLen(element_name) == 0) {
         // Corresponds to #PCDATA in a mixed content
         return new RegExpr::Node(RegExpr::EPSILON);
      } else {
         return new RegExpr::Node(RegExpr::LABEL,
                                  alphabet->find_id(getSymbol(element_name)));
      }
      break;
   case ContentSpecNode::Sequence :
      return new RegExpr::Node(RegExpr::CONCAT,
                               translate(spec->getFirst(),alphabet),
                               translate(spec->getSecond(),alphabet));
      break;
   case ContentSpecNode::Choice :
      return new RegExpr::Node(RegExpr::UNION,
                               translate(spec->getFirst(),alphabet),
                               translate(spec->getSecond(),alphabet));
      break;
   case ContentSpecNode::ZeroOrOne :
      return new RegExpr::Node(RegExpr::UNION,
                               new RegExpr::Node(RegExpr::EPSILON),
                               translate(spec->getFirst(),alphabet));
      break;
   case ContentSpecNode::OneOrMore :
      child = translate(spec->getFirst(),alphabet);
      return new RegExpr::Node(RegExpr::CONCAT,
                               child, 
                               new RegExpr::Node(RegExpr::STAR,child,NULL));
      break;
   case ContentSpecNode::ZeroOrMore :
      return new RegExpr::Node(RegExpr::STAR,
                               translate(spec->getFirst(),alphabet),
                               NULL);
      break;
   default:
      ERROR("Unhandled content specification type %i",spec->getType());
      break;
   }
   return NULL;
}

void Validator::gatherAndTranslateSymbols(const ContentSpecNode* spec, 
                                          List<String>& symbols)
{
   const XMLCh* element_name;

   switch (spec->getType()) {
   case ContentSpecNode::Leaf :
      element_name = spec->getElement()->getRawName();
      if (XMLString::stringLen(element_name) != 0) {
         symbols.add(getSymbol(element_name));
      }
      break;
   case ContentSpecNode::Choice :
      gatherAndTranslateSymbols(spec->getFirst(),symbols);
      gatherAndTranslateSymbols(spec->getSecond(),symbols);
      break;
   case ContentSpecNode::ZeroOrOne :
      gatherAndTranslateSymbols(spec->getFirst(),symbols);
      break;
   case ContentSpecNode::OneOrMore :
      gatherAndTranslateSymbols(spec->getFirst(),symbols);
      break;
   case ContentSpecNode::ZeroOrMore :
      gatherAndTranslateSymbols(spec->getFirst(),symbols);
      break;
   case ContentSpecNode::Sequence :
   default:
      ERROR("Unhandled content specification type %i",spec->getType());
      break;
   }
}

void Validator::gatherAndTranslateSymbols(const ContentSpecNode* spec, 
                                          Vector<String>& symbols)
{
   List<String> l;

   gatherAndTranslateSymbols(spec,l);
   symbols.reinit(l);
}

RegExpr* Validator::translateContentSpec(const ContentSpecNode* spec)
{
   Alphabet* alphabet = new Alphabet();
   return new RegExpr(alphabet,translate(spec,alphabet));
}

bool Validator::checkContent(XMLElementDecl* decl, 
                             QName** children, 
                             XMLSize_t count, 
                             XMLSize_t* indexFailingChild)
{
   
   RegExpr* e;
   Vector<String> symbols;

   //Getting the name of the element. 
   const XMLCh* element_name = decl->getFullName();

   // Fetching the proper recognizer.
   Recognizer* r = recognizer[element_name];
   DTDElementDecl* d = (DTDElementDecl*) decl;

   //Building the appropriate recognizer if one doesn't yet exits.
   if (r == NULL) {
      switch (d->getModelType()) {
      case DTDElementDecl::Empty :
         r = Recognizer::EMPTY_WORD_ONLY;
         break;
      case DTDElementDecl::Any :
         r = Recognizer::UNIVERSAL;
         break;
      case DTDElementDecl::Mixed_Simple :
         DEBUG("Mixed content");
         gatherAndTranslateSymbols(d->getContentSpec(),symbols);
         preprocessing->start();
         r = new ClosureRecognizer(symbols);
         preprocessing->stop();
         break;
      case DTDElementDecl::Children :
         DEBUG("In Children content");
         e = translateContentSpec(d->getContentSpec());
         preprocessing->start();
         r = f->createRecognizer(e);
         preprocessing->stop();
         DEBUG("Finished building a recognizer"); 
         break;
      case DTDElementDecl::ModelTypes_Count :
         ERROR("Unhandled DTD model type 'ModelTypes_Count'");
      }
      recognizer[element_name] = r;
   }

   // Translating the children labels vector
   Vector<String> symbol(count);
   for (XMLSize_t i = 0; i < count; i++) {
      symbol[i] = getSymbol(children[i]->getRawName());
   }

   matching->start();
   bool result = r->match(symbol);
   matching->stop();

   return result;
}


void Validator::faultInAttr(XMLAttr& attr, const XMLAttDef& attr_def) const
{
}

void Validator::preContentValidation(bool, bool)
{
}

void Validator::postParseValidation()
{
}

void Validator::reset()
{
}

bool Validator::requiresNamespaces() const
{
   return false;
}

void Validator::validateAttrValue(const XMLAttDef*, const XMLCh*, bool, const XMLElementDecl*)
{
}

void Validator::validateElement(const XMLElementDecl* decl)
{
}

Grammar* Validator::getGrammar() const
{
   return this->grammar;
}

void Validator::setGrammar(Grammar* grammar)
{
      this->grammar = grammar;
}

bool Validator::handlesDTD() const
{
   return true;
}

bool Validator::handlesSchema() const
{
   return false;
}
