//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_VEB_H_
#define LiRE_VEB_H_

#include "vector.h"
#include "array.h"

template<class T> 
struct OrdMap
{
   int domain;    // the size of the universe.
   int size;      // the number of elements in the structure. 
   int min;       // the minimum value in the set.  
   int max;       // the maximum value in the set.
   Vector<T> a;

   OrdMap(int domain);
   ~OrdMap();
   void insert(int i, T n);
   T get(int i);
   int  find_succ(int i);
   int  find_pred(int i);
   bool contains(int i);
};

template<class T> 
OrdMap<T>::OrdMap(int domain) : a(domain)
{
   this->domain = domain;
   for (int i = 0; i < a.size; i++) {
      this->a[i] = T();
   }
   this->size = 0;
   this->min = domain;
   this->max = -1;
}

template<class T> 
OrdMap<T>::~OrdMap()
{
}

template<class T> 
void OrdMap<T>::insert(int k, T v)
{
   if (! this->contains(k)) {
      this->size++;
      if (k < this->min) {
         this->min = k;
      } else if (k > this->max) {
         this->max = k;
      }
   }

   this->a[k] = v;
}

template<class T> 
T OrdMap<T>::get(int k)
{
   if (k < 0 || k >= this->domain) {
      return T();
   }

   return this->a[k];
}

template<class T> 
int OrdMap<T>::find_pred(int k)
{
   if (!( 0 <= k && k < this->domain)) {
      return -1;
   }

   for (int i = k; i >= 0; i--) {
      if (this->a[i]) {
         return i;
      }
   }

   return -1;
}

template<class T> 
int OrdMap<T>::find_succ(int k)
{
   if (!( 0 <= k && k < this->domain)) {
      return -1;
   }

   for (int i = k; i < domain; i++) {
      if (this->a[i]) {
         return i;
      }
   }

   return -1;
}

template<class T> 
bool OrdMap<T>::contains(int i)
{
   return this->a[i];
}

// The van Emde Boas structure. 
template<class T> 
struct VEB : OrdMap<T>
{
   // int domain;    // the size of the universe.
   // int size;      // the number of elements in the structure. 
   // int min;       // the minimum value in the set. (this value is not stored recursively).  
   // int max;       // the maximum value in the set.
   
   T value;  // the value assotiated with the minimum (if min != domain+1)

   int sqrt_domain; // 'square root' of the number above.

   int high(int i) { return i / this->sqrt_domain; };
   int low(int i) { return i % this->sqrt_domain; };
   // int high(int i) { return i / sqrt_domain; };
   // int low(int i) { return i % sqrt_domain; };

   VEB* top;            // The top structure; NULL by default.
   HArray<VEB*> bottom; // The bottom structure; NULL by default.

   VEB(int domain);
   ~VEB();
   void insert(int i, T n);
   T get(int i);
   int find_succ(int i);
   int find_pred(int i);
   bool contains(int i);
};

template<class T> 
VEB<T>::VEB(int domain) : OrdMap<T>(0),
   bottom((domain > 1) ? (sqrt_ref((unsigned long long)domain)) : (0))
{
   ASSERT(domain > 0);

   this->domain = domain;
   this->sqrt_domain = sqrt_ref((unsigned long long)domain); 
   this->min = domain;
   this->max = -1;
   this->value = T();
   this->top = NULL;
   this->size = 0;
}

template<class T> 
VEB<T>::~VEB()
{
   if (this->top != NULL) {
      delete this->top;
   }

   for (int i = 0; i < this->bottom.size; i++) {
      if (this->bottom[i] != NULL) {
         delete this->bottom[i];
      }
   }
   
}

template<class T> 
void VEB<T>::insert(int i, T v)
{
   // The new key is new (not present in the set).
   ASSERT(0 <= i && i < this->domain);
   ASSERT(! this->contains(i));


   if (this->size == 0) {
      this->min = i;
      this->max = i;
      this->value = v;
   } else {
      // size != 0 and !contains(i) imply that domain > 1;
      if (i < this->min){
         swap(i,this->min);
         swap(v,this->value); 
      } 
      // i = a*sqrt(domain) + b
      int a = this->high(i);  // a = i / sqrt(domain);
      int b = this->low(i);   // b = i % sqrt(domain);

      if (this->top == NULL) {
         this->top = new VEB(this->sqrt_domain);
      }
      if (this->bottom[a] == NULL) {
         this->bottom[a] = new VEB(this->sqrt_domain);
      }

      if (this->bottom[a]->size == 0) {
         this->top->insert(a,T());
      }

      // if bottom[a].size == 0, then the following operation runs in
      // O(1) time.
      // if (i == 281 && sqrt_domain == 24) {
      //    printf("hello");
      // }
      this->bottom[a]->insert(b,v); 
      
      if (this->max < i) {
         this->max = i;
      }
   }

   this->size++;
}

template<class T> 
T VEB<T>::get(int i)
{

   if (this->size == 0 || 
       !( 0 <= i && i < this->domain)) {
      return T();
   }

   if (i < this->min || this->max < i) {
      return T();
   }

   if (i == this->min) {
      return this->value;
   }

   // since size > 0, top has been already initialized. 
   ASSERT(this->top != NULL);

   // i = a*sqrt(domain) + b
   int a = this->high(i);  // a = i / sqrt(domain);
   int b = this->low(i);   // b = i % sqrt(domain);

   if (this->bottom[a] == NULL) {
      return T();
   } else {
      return this->bottom[a]->get(b);
   }
}

template<class T> 
int VEB<T>::find_succ(int i)
{
   if (this->size == 0 || 
       !(0 <= i && i < this->domain)) {
      return -1;
   }

   if (this->max < i) {
      return -1;
   }

   if (i <= this->min) {
      return this->min;
   }
   // since size > 0, top has been already initialized. 
   ASSERT(this->top != NULL);

   // i = a*sqrt(domain) + b
   int a = this->high(i);  // a = i / sqrt(domain);
   int b = this->low(i);   // b = i % sqrt(domain);

   int res;

   if (this->bottom[a] != NULL && 
       this->bottom[a]->max >= b) {
      res = a * this->sqrt_domain + 
         this->bottom[a]->find_succ(b);
   } else {
      // a+1 < sqrt_domain from the fact that i <= max.
      int c = this->top->find_succ(a+1);
      if (c == -1) {
         return -1;
      }
      res = c * this->sqrt_domain + this->bottom[c]->min;
   }

   return res;
}

template<class T> 
int VEB<T>::find_pred(int i)
{
   if (this->size == 0 || 
       !( 0 <= i && i < this->domain)) {
      return -1;
   }

   if (i < this->min) {
      return -1;
   } 

   if (this->min == i) {
      return this->min;
   }

   if (this->max <= i) {
      return this->max;
   }

   // since size > 0, top has been already initialized. 
   ASSERT(this->top != NULL);

   // i = a*sqrt(domain) + b
   int a = this->high(i);  // a = i / sqrt(domain);
   int b = this->low(i);   // b = i % sqrt(domain);

   int res;

   if (this->bottom[a] != NULL && 
       this->bottom[a]->min <= b) {
      res = a * this->sqrt_domain + 
         this->bottom[a]->find_pred(b);
   } else {
      // bottom[a] is empty but min <= i
      // a+1 < sqrt_domain from the fact that i <= max.
      int c = this->top->find_pred(a-1);
      if (c == -1) {
         return this->min;
      }
      res = c * this->sqrt_domain + this->bottom[c]->max;
   }

   return res;
}

template<class T> 
bool VEB<T>::contains(int i) 
{ 
   return this->find_succ(i) == i; 
}

#endif // LiRE_VEB_H_
