//
// This file is part of LiRE, a free, experimental library for
// efficient testing of determinism and membership for regular
// expressions.
//
// LiRE is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, version 3 of the License, or any
// later version.
//
// LiRE is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
// License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LiRE. If not, see <http://www.gnu.org/licenses/>.
//
// Authors: Sławek Staworko <staworko@gmail.com>
//          Benoît Groz <benoit.groz@yahoo.fr>
//          Sebastian Maneth <sebastian.maneth@gmail.com>
// Date: 20 June 2011.
//
#ifndef LiRE_VECTOR_H_
#define LiRE_VECTOR_H_

#include <stdlib.h>

#include "list.h"
#include "error.h"
#include "utils.h"
#include "logger.h"

//
// VECTOR. Fixed-length array wrapper with reference counter and
// subvector facility.
//
template<class T> 
struct Vector 
{
   T*   res;
   int* ref_count;

   T*   a;
   int  size;

   Vector();
   Vector(int size);
   Vector(int size, T val);
   Vector(const Vector<T>& v);
   Vector(const Vector<T>& v, int offset, int size);
   Vector(const List<T>& l);

   ~Vector();

   void init();
   void init(int size);
   void init(int size, T val);
   void init(const Vector<T>& l);
   void init(const Vector<T>& l, int offset, int size);
   void init(const List<T>& l);

   void reinit();
   void reinit(int size);
   void reinit(int size, T val);
   void reinit(const Vector<T>& l);
   void reinit(const Vector<T>& l, int offset, int size);
   void reinit(const List<T>& l);

   void purge();

   void fill(T val);

   bool null() const;
   bool empty() const;

   T& operator[](int i);

   // Disable assignment and copy;
   void operator=(const Vector& l);
   void operator=(const List<T>& l);
   void operator=(int size);
};

template<class T> 
Vector<T>::Vector()
{
   init();
}

template<class T> 
Vector<T>::Vector(int size)
{
   init(size);
}

template<class T> 
Vector<T>::Vector(int size, T v)
{
   init(size,v);
}

template<class T> 
Vector<T>::Vector(const List<T>& l)
{
   init(l);
}

template<class T> 
Vector<T>::Vector(const Vector<T>& v)
{
   init(v);
}

template<class T> 
Vector<T>::Vector(const Vector<T>& v, int offset, int size)
{
   init(v,offset,size);
}

template<class T> 
Vector<T>::~Vector()
{
   this->purge();
}

template<class T> 
void Vector<T>::init() 
{
   this->res = NULL;
   this->ref_count = NULL;
   this->a = NULL;
   this->size = -1;
}

template<class T> 
void Vector<T>::init(int size)
{
   ASSERT(size >= 0);

   this->res = new T[size];
   this->ref_count = new int;
   *(this->ref_count) = 1;

   this->a = this->res;
   this->size = size;
}

template<class T> 
void Vector<T>::init(int size, T val)
{
   init(size);
   fill(val);      
}

template<class T> 
void Vector<T>::init(const Vector<T>& v)
{
   this->res = v.res;
   this->ref_count = v.ref_count;
   if (this->ref_count != NULL) {
      *(this->ref_count) = *(this->ref_count) + 1;
   }

   this->a = v.a;
   this->size = v.size;
}

template<class T> 
void Vector<T>::init(const Vector<T>& v, int offset, int size)
{
   ASSERT(!v.null());
   ASSERT(offset >= 0);
   ASSERT(size >= 0);
   ASSERT(offset <= v.size);
   ASSERT(offset + size <= v.size);

   this->res = v.res;
   this->ref_count = v.ref_count;
   *(this->ref_count) = *(this->ref_count) + 1;

   this->a = v.a + offset;
   this->size = size;
}

template<class T> 
void Vector<T>::init(const List<T>& l) 
{
   init();
   l.pack_into(*this);
}

template<class T> 
void Vector<T>::reinit() 
{
   purge();
   init();
}

template<class T> 
void Vector<T>::reinit(int size) 
{
   purge();
   init(size);
}

template<class T> 
void Vector<T>::reinit(int size, T val)
{
   purge();
   init(size,val);
}

template<class T> 
void Vector<T>::reinit(const Vector<T>& v) 
{
   purge();
   init(v);
}

template<class T> 
void Vector<T>::reinit(const Vector<T>& v, int offset, int size) 
{
   purge();
   init(v,offset,size);
}

template<class T> 
void Vector<T>::reinit(const List<T>& l) 
{
   purge();
   init(l);
}

template<class T> 
void Vector<T>::purge()
{
   if (this->res != NULL) {
      this->res = NULL;
      *(this->ref_count) = *(this->ref_count) - 1;
      if (*(this->ref_count) == 0) {
         delete[] this->res;
         delete this->ref_count;
      } 
      this->a = NULL;
      this->ref_count = NULL;
      this->size = -1;
   }
}

template<class T> bool Vector<T>::null() const
{
   return this->res == NULL;
}

template<class T> 
bool Vector<T>::empty() const
{
   return this->null() || this->size == 0;
}

template<class T> 
void Vector<T>::fill(T val)
{
   for (int i = 0; i < size; i++) {
      (*this)[i] = val;
   }
}

template<class T> 
T& Vector<T>::operator[](int i)
{
   ASSERT(this->a != NULL);
   ASSERT(0 <= i, "Range bound error.");
   ASSERT(i < this->size, "Range bound error.");
   return this->a[i];
}


template<class T>
static inline void swap(Vector<T>& a, Vector<T>& b)
{
   swap(a.res,b.res);
   swap(a.ref_count,b.ref_count);
   swap(a.a,b.a);
   swap(a.size,b.size);
}

#endif // LiRE_VECTOR_H_
